from warnings import filterwarnings, catch_warnings
import numpy as np
import pandas as pd
from scipy.ndimage.measurements import mean, standard_deviation, center_of_mass


class Measurements:
    """
    Object measures properties of labeled segments within an image.

    Attributes:

        colordepth (int) - number of color channels

        segment_ids (np.ndarray[float]) - ordered segment labels

        levels (dict) - {channel: np.ndarray[float]} - expression levels

        std (dict) - {channel: np.ndarray[float]} - expression std. deviation

        xpos (np.ndarray[float]) - segment centroid x-positions

        ypos (np.ndarray[float]) - segment centroid y-positions

        voxel_size (np.ndarray[float]) - segment voxel size

    """

    def __init__(self, im, labels):
        """
        Measure properties of labeled segments within an image.

        Args:

            im (np.ndarray[float]) - 3D array of pixel values

            labels (np.ndarray[int]) - cell segment labels

        """

        self.colordepth = im.shape[-1]

        # set segment ids (ordered)
        self.segment_ids = np.unique(labels[labels.nonzero()])

        # measure expression levels
        self.measure_expression(im, labels, self.segment_ids)

        # measure segment centroids
        self.measure_centroids(labels, self.segment_ids)

        # measure segment voxel sizes
        self.measure_segment_size(labels, self.segment_ids)

    def measure_expression(self, im, labels, segment_ids):
        """
        Measure expression levels.

        Args:

            im (np.ndarray[float]) - 3D array of pixel values

            labels (np.ndarray[int]) - cell segment labels

            segment_ids (np.ndarray[int]) - ordered segment IDs

        """

        # split R/G/B image channels
        drop = lambda x: x.reshape(*x.shape[:2])
        channels = [drop(x) for x in np.split(im, self.colordepth, axis=-1)]

        # compute means
        means = [mean(channel, labels, segment_ids) for channel in channels]

        # compute std
        with catch_warnings():
            filterwarnings('ignore')
            evaluate_std = lambda x: standard_deviation(x, labels, segment_ids)
            stds = [evaluate_std(channel) for channel in channels]

        # compile dictionaries
        self.levels = dict(enumerate(means))
        self.std = dict(enumerate(stds))

    def measure_centroids(self, labels, segment_ids):
        """
        Measure the centroid of each segment.

        Args:

            labels (np.ndarray[int]) - cell segment labels

            segment_ids (np.ndarray[int]) - ordered segment IDs

        """
        centroid_dict = self.evaluate_centroids(labels)
        centroids = [centroid_dict[seg_id] for seg_id in segment_ids]
        xpos, ypos = list(zip(*centroids))
        self.xpos = xpos
        self.ypos = ypos

    @staticmethod
    def evaluate_centroids(labels):
        """
        Evaluate center of mass of each label.

        * Note: scipy returns centroids as (y, x) which are flipped to (x, y)

        Args:

            labels (np.ndarray[int]) - segment label mask

        Returns:

            center_of_mass (dict) - {segment_id: [xpos, ypos]} pairs

        """

        seg_ids = np.unique(labels[labels!=0])
        coms = center_of_mass(labels, labels, seg_ids)
        return {seg_id: com[::-1] for seg_id, com in zip(seg_ids, coms)}

    def measure_segment_size(self, labels, segment_ids):
        """
        Measure the voxel size of each segment.

        Args:
        labels (np.ndarray[int]) - cell segment labels
        segment_ids (np.ndarray[int]) - ordered segment IDs
        """
        voxels = labels[labels!=0]
        bins = np.arange(0, segment_ids.max()+3, 1)
        counts, _ = np.histogram(voxels, bins=bins)
        self.voxel_counts = counts[segment_ids]

    def build_dataframe(self):
        """
        Build and return dataframe containing all measurements.

        Returns:

            measurements (pd.DataFrame) - measurement data

        """

        # construct dataframe
        measurement_data = dict(
            segment_id=self.segment_ids,
            centroid_x=self.xpos,
            centroid_y=self.ypos,
            pixel_count=self.voxel_counts)

        # add intensity measurements
        for channel_id in range(self.colordepth):

            # define keys
            key = 'ch{:d}'.format(channel_id)
            key_std = key + '_std'

            # store measured levels
            measurement_data[key] = self.levels[channel_id]
            measurement_data[key_std] = self.std[channel_id]

        return pd.DataFrame(measurement_data)



import numpy as np
from sklearn.neighbors import KernelDensity
from sklearn.cluster import k_means


class KDE:
    """
    Kernel density fit to node coordinates.
    """

    def __init__(self, xy, bandwidth=100, n=2):
        """
        Instantiate KDE object.

        Args:

            xy (2D np.ndarray[float]) - node positions

            bandwidth (float) - bandwidth for KernelDensity

            n (int) - number of clusters

        """

        self.xy = xy
        self.bandwidth = bandwidth
        self.kde = self.fit_kde(self.xy, bandwidth)
        self.density = self.evaluate_density()
        self.cluster_density(n)
        self.evaluate_cluster_threshold()

    def __call__(self, x):
        return np.exp(self.kde.score_samples(x))

    @staticmethod
    def fit_kde(xy, bandwidth=100):
        """ Fit gaussian kernel to xy coordinates spatial density."""
        kde = KernelDensity(bandwidth=bandwidth, kernel='gaussian')
        kde.fit(xy)
        return kde

    def evaluate_density(self):
        """ Returns density estimate for all cell measurement positions. """
        return self.__call__(self.xy)

    def cluster_density(self, n=2):
        values = np.log10(self.density.reshape(-1, 1))
        centroid, label, inertia = k_means(values, n)
        self.centroid = centroid
        self.label = label

    def evaluate_cluster_threshold(self, cluster_id=0):
        order = np.argsort(self.centroid.flatten())
        threshold = self.density[self.label==order[0]].max()
        self.threshold = threshold
        self.mask = self.density > threshold



from itertools import combinations
from collections import OrderedDict
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from scipy.stats import ks_2samp, ttest_ind, mannwhitneyu

from ..visualization import *


class PairwiseComparison:
    """
    Pairwise statistical comparison between measurements of two populations.

    Attributes:

        x (pd.DataFrame) - measurements of first population

        y (pd.DataFrame) - measurements of second population

        basis (str) - attribute on which populations are compared

    """

    def __init__(self, x, y, basis):
        """
        Instantiate pairwise comparison of two populations.

        Attributes:

            x (pd.DataFrame) - measurements of first population

            y (pd.DataFrame) - measurements of second population

            basis (str) - attribute on which populations are compared

            """
        self.x = x
        self.y = y
        self.basis = basis

    @property
    def is_greater_than(self):
        """ True if first population mean is greater than the second."""
        return self.x[self.basis].mean() > self.y[self.basis].mean()

    def compare(self, test='MW', **kwargs):
        """
        Run statistical test.

        Args:

            test (str) - name of test used, one of ('KS', 't', 'MW')

        Returns:

            p (float) - p-value

            is_greater_than (bool) - True if first population mean is greater

            kwargs: keyword arguments for statistical test

        """

        # extract compared values for each population
        x, y = self.x[self.basis], self.y[self.basis]

        # perform statistical test
        if test.lower() == 'ks':
            k, p = ks_2samp(x, y, **kwargs)
        elif test.lower() == 't':
            k, p = ttest_ind(x, y, **kwargs)
        elif test.lower() == 'mw':
            k, p = mannwhitneyu(x, y, **kwargs)
        else:
            raise ValueError('Test {:s} not recognized.'.format(test))

        return p, self.is_greater_than


class PairwiseCelltypeComparison(PairwiseComparison):
    """
    Pairwise statistical comparison between two concurrent cell types.

    Attributes:

        label (str) - attribute used to stratify populations

        type1 (str) - first label

        type2 (str) - second label

    Inherited attributes:

        x (pd.DataFrame) - measurements of first population

        y (pd.DataFrame) - measurements of second population

        basis (str) - attribute on which populations are compared

    """

    def __init__(self, measurements, type1, type2, basis,
                 label='celltype',
                 concurrent_only=True):
        """
        Instantiate comparison between two concurrent cell types.

        Args:

            measurements (pd.DataFrame) - measurement data

            type1 (str or int) - first label

            type2 (str or int) - second label

            basis (str) - attribute on which populations are compared

            label (str) - attribute used to define population labels

            concurrent_only (bool) - if True, only compare concurrent cells

        """

        # store labels
        self.label = label
        self.type1 = type1
        self.type2 = type2

        # select concurrent cells of each type
        if concurrent_only:
            k1 = 'concurrent_'+str(self.type1)
            k2 = 'concurrent_'+str(self.type2)
            measurements = measurements[measurements[k1] & measurements[k2]]

        # split into two populations
        x = measurements[measurements[label] == self.type1]
        y = measurements[measurements[label] == self.type2]

        # instantiate comparison
        super().__init__(x, y, basis=basis)

    def plot(self,
             ax=None,
             colors=None,
             mode='violin',
             ylabel=None,
             **kwargs):
        """
        Visualize comparison using seaborn box or violinplot.

        Args:

            ax (matplotlib.axes.AxesSubplot)

            colors (dict) - color for each box/violin keyed by label

            mode (str) - type of comparison, either 'box', 'violin', or 'strip'

            ylabel (str) - label for yaxis

            kwargs: keyword arguments for seaborn plotting function

        """

        # create figure if none provided
        if ax is None:
            fig, ax = plt.subplots(figsize=(1.25, 1.5))

        # plot boxplot
        if mode == 'box':
            sns.boxplot(ax=ax,
                    x=self.label,
                    y=self.basis,
                    data=pd.concat((self.x, self.y)),
                    order=(self.type1, self.type2),
                    notch=True,
                    width=0.8,
                    **kwargs)

        # plot violinplot
        elif mode == 'violin':
            sns.violinplot(ax=ax,
                    x=self.label,
                    y=self.basis,
                    data=pd.concat((self.x, self.y)),
                    order=(self.type1, self.type2),
                    scale='width',
                    linewidth=0.5,
                    **kwargs)

        # plot stripplot
        elif mode == 'strip':
            sns.stripplot(ax=ax,
                    x=self.label,
                    y=self.basis,
                    data=pd.concat((self.x, self.y)),
                    order=(self.type1, self.type2),
                    dodge=True,
                    **kwargs)

        # format axis
        self.format_axis(ax, colors=colors, mode=mode, ylabel=ylabel)

    def format_axis(self, ax,
                    colors=None,
                    axis_labels=None,
                    mode='violin',
                    ylabel=None):
        """
        Format axis.

        Args:

            ax (matplotlib.axes.AxesSubplot)

            colors (dict) - color for each box/violin keyed by label

            axis_labels (dict) - axis label for each box/violin keyed by label

            mode (str) - type of comparison, either 'box', 'violin', or 'strip'

            ylabel (str) - label for y axis

        """

        # format axes
        ax.grid(False)
        ax.yaxis.set_ticks_position('left')
        ax.xaxis.set_ticks_position('bottom')
        _ = ax.spines['top'].set_visible(False)
        _ = ax.spines['right'].set_visible(False)

        # get violin artists
        if mode == 'violin':
            is_poly = lambda x: x.__class__.__name__ == 'PolyCollection'
            polys = [c for c in ax.collections if is_poly(c)]

        # format xticks
        ticklabels = []

        for i, label in enumerate(ax.get_xticklabels()):

            # set violin/box color
            if colors is not None:
                if mode == 'violin':
                    polys[i].set_color(colors[label.get_text()[0]])
                else:
                    ax.artists[i].set_facecolor(colors[label.get_text()[0]])

            # set tick label as celltype
            if axis_labels is not None:
                label.set_text(axis_labels[label.get_text()[0]])
                ticklabels.append(label)

        # format xlabels
        ax.set_xlabel('')
        if axis_labels is not None:
            _ = ax.set_xticklabels(ticklabels, ha='center')

        # set ylabel
        if ylabel is not None:
            ax.set_ylabel(ylabel)


class CelltypeComparison:
    """
    Summary of comparisons between all labeled celltypes.

    Attributes:

        measurements (pd.DataFrame) - measurement data

        basis (str) - attribute on which populations are compared

        label (str) - attribute used to define population labels

        test (str) - statistical test used, one of ('KS', 't', 'MW')

    """

    def __init__(self, measurements, basis,
                 label='celltype', test='MW', **kwargs):
        """
        Instantiate summary of comparisons between all labeled cell types.

        Args:

            measurements (pd.DataFrame) - measurement data

            basis (str) - attribute on which populations are compared

            label (str) - attribute used to define population labels

            test (str) - name of test used, one of ('KS', 't', 'MW')

            kwargs: keyword arguments for statistical test

        """
        self.measurements = measurements
        self.basis = basis
        self.label = label
        self.test = test

        # compute and report pvalues
        pvals = self.run(**kwargs)
        self.report(pvals)

    @property
    def pairs(self):
        """ Unique pairs of labels. """
        label_values = self.measurements[self.label].unique()
        return list(sorted([sorted(x) for x in combinations(label_values, 2)]))

    def compare_celltype(self, type1, type2, **kwargs):
        """
        Args:

            type1 (str) - first cell type

            type2 (str) - second cell type

            kwargs: keyword arguments for statistical test

        Returns:

            p (float) - p value for comparison statistic

            is_greater_than (bool) - True if first population mean is greater

        """
        comparison = PairwiseCelltypeComparison(self.measurements, type1, type2, self.basis, label=self.label)
        return comparison.compare(self.test, **kwargs)

    def run(self, **kwargs):
        """
        Compare all pairwise combinations of cell types.

        kwargs: keyword arguments for statistical test

        Returns:

            pvals (dict) - {comparison: pvalue} pairs

        """

        signs = {True: ' > ', False: ' < '}

        pvals = OrderedDict()
        for pair in self.pairs:
            pval, greater_than = self.compare_celltype(*pair, **kwargs)
            comparison_name = signs[greater_than].join([str(x) for x in pair])
            pvals[comparison_name] = pval

        return pvals

    def report(self, pvals):
        """
        Print summary of statistical comparisons for each condition.

        Args:

            pvals (dict) - {comparison: pvalue} pairs

        """
        print('Statistical test: {}'.format(self.test))
        for test, pval in pvals.items():
            print(test + ': p = {:0.4f}'.format(pval))


# # define labels and corresponding fill colors
# axis_labels = {'m': '−/−', 'h': '−/+', 'w': '+/+',
#           '0': '−/−', '1': '−/+', '2': '+/+'}

# if colors is None:
#     colors = {'m': '−/−', 'h': '−/+', 'w': '+/+',
#               '0': 'y', '1': 'm', '2': 'c'}



import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec


def default_figure(func):
    """ Decorator for creating axis. """
    def wrapper(*args, ax=None, figsize=(2., 1.25), **kwargs):
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
            ax.spines['top'].set_visible(False)
            ax.spines['right'].set_visible(False)
        output = func(*args, ax=ax, **kwargs)
        if output is None:
            output = plt.gcf()
        return output
    return wrapper


def square_figure(func):
    """ Decorator for creating square axis without spines. """
    def wrapper(*args, ax=None, figsize=(2., 2.), **kwargs):
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
            ax.spines['top'].set_visible(False)
            ax.spines['right'].set_visible(False)
        output = func(*args, ax=ax, **kwargs)
        ax.set_aspect(1)
        ax.axis('off')
        if output is None:
            output = plt.gcf()
        return output
    return wrapper


def joint_figure(func):
    """ Decorator for creating joint distribution figure. """

    def format_joint_axis(ax):
        """ Format joint axis. """
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.set_xlabel('ln X')
        ax.set_ylabel('<ln X> among neighbors')
        ax.spines['left'].set_position(('outward', 2))
        ax.spines['bottom'].set_position(('outward', 2))

    def wrapper(self, *args, size_ratio=4, figsize=(2., 2.), **kwargs):

        # create figure
        fig = plt.figure(figsize=figsize)
        ratios = [size_ratio/(1+size_ratio), 1/(1+size_ratio)]
        gs = GridSpec(2, 2, width_ratios=ratios,
                      height_ratios=ratios[::-1], wspace=0, hspace=0)
        fig.ax_joint = fig.add_subplot(gs[1, 0])
        fig.ax_xmargin = fig.add_subplot(gs[0, 0])
        fig.ax_ymargin = fig.add_subplot(gs[1, 1])
        fig.ax_xmargin.axis('off')
        fig.ax_ymargin.axis('off')
        format_joint_axis(fig.ax_joint)

        # run plotting function
        output = func(self, fig, *args, **kwargs)

        # invert yaxis (after all plotting is done)
        fig.ax_joint.invert_yaxis()

        # set axis limits
        fig.ax_joint.set_xlim(fig.ax_joint.get_ylim())
        fig.ax_xmargin.set_xlim(fig.ax_joint.get_ylim())

        return output

    return wrapper



from os.path import join, abspath, relpath, isdir, exists
from os import mkdir, chmod, pardir
import shutil

import numpy as np
import pandas as pd
import dill as pickle

from ..utilities import Pickler
from .arguments import str2bool

from growth.sweep.sweep import Sweep
from .batch import BatchBenchmark
from .visualization import SweepVisualization
from .results import BenchmarkingResults
from .growth import GrowthTrends


class SweepBenchmark(Pickler, SweepVisualization):

    """
    Attributes:

        sweep_path (str) - path to parent growth.sweep.Sweep directory

        path (str) - path to benchmark subdirectory

        batches (2D np.ndarray[growth.Batch]) - batches of growth replicates

        ambiguities (np.ndarray[float]) - fluorescence ambiguities

        num_replicates (int) - number of fluorescence replicates

        script_name (str) - name of run script

    """

    def __init__(self, sweep_path,
                 min_ambiguity=0.1,
                 max_ambiguity=1.,
                 num_ambiguities=16,
                 num_replicates=1,
                 script_name='run_batch.py'):

        # load and set growth sweep
        self.sweep_path = sweep_path
        self.batches = Sweep.load(sweep_path).batches

        # set fluorescence ambiguities
        self.ambiguities = np.linspace(min_ambiguity, max_ambiguity, num_ambiguities)

        # set number of fluorescence replicates
        self.num_replicates = num_replicates

        # instantiate dictionary of paths to each BatchBenchmark object
        self.benchmark_paths = {}

        # set script name
        self.script_name = script_name

        # initialize results
        self._results = None

    @property
    def sweep(self):
        """ Sweep of growth replicates. """
        return Sweep.load(self.sweep_path)

    @property
    def trends(self):
        """ Growth trends. """
        return GrowthTrends(self.sweep._results)

    @property
    def mean_clone_sizes(self):
        """ Mean size of clones for each growth condition. """
        return np.log(self.trends.means.mean_clone_size.values)

    @staticmethod
    def load(sweep_path):
        """
        Load job.

        Args:

            sweep_path (str) - path to growth.Sweep directory

        """
        path = join(sweep_path, 'benchmark')
        with open(join(path, 'benchmark_job.pkl'), 'rb') as file:
            job = pickle.load(file)

        # set path
        job.sweep_path = sweep_path
        job.path = path
        job.batches = Sweep.load(sweep_path).batches

        # load results
        results_path = join(path, 'data.hdf')
        if exists(results_path):
            try:
                job._results = pd.read_hdf(results_path, 'benchmark')
            except:
                job._results = None
        return job

    @property
    def num_ambiguities(self):
        """ Number of fluorescence ambiguities. """
        return self.ambiguities.size

    @staticmethod
    def build_run_script(path, python_script_name,
                         shell_script_name='run.sh',
                         **kwargs):
        """
        Writes bash run script for local use.

        Args:

            path (str) - path to benchmarking top directory

            python_script_name (str) - name of python script

            shell_script_name (str) - name of shell script

            kwargs: keyword arguments used at runtime

        """

        # define paths
        path = abspath(path)
        run_path = path.rsplit('/', maxsplit=1)[0]
        job_script_path = join(path, 'scripts', shell_script_name)

        # copy run script to scripts directory
        scripts_path = abspath(__file__).rsplit('/', maxsplit=2)[0]
        run_script = join(scripts_path, 'scripts', python_script_name)
        shutil.copy(run_script, join(path, 'scripts'))

        # compile runtime command
        cmd = 'python ./benchmark/scripts/{:s}'.format(python_script_name)+' ${P}'
        for k, v in kwargs.items():
            cmd += ' -{} {}'.format(k, v)

        # declare outer script that reads PATH from file
        job_script = open(job_script_path, 'w')
        job_script.write('#!/bin/bash\n')

        # move to job directory
        job_script.write('cd {:s} \n\n'.format(run_path))

        # run each batch
        job_script.write('echo "Starting all batches at `date`"\n')
        job_script.write('while read P; do\n')
        job_script.write('echo "Processing batch ${P}"\n')
        job_script.write(cmd + ' \n')
        job_script.write('done < ./benchmark/jobs/index.txt \n')
        job_script.write('echo "Job completed at `date`"\n')
        job_script.write('exit\n')

        # close the file
        job_script.close()

        # change the permissions
        chmod(job_script_path, 0o755)

    @staticmethod
    def build_submission_script(path,
                                script_name,
                                walltime=10,
                                allocation='p30653',
                                cores=1,
                                memory=4,
                                **kwargs):
        """
        Writes job submission script for QUEST.

        Args:

            path (str) - path to benchmarking top directory

            script_name (str) - name of run script

            walltime (int) - estimated job run time

            allocation (str) - project allocation, e.g. p30653 (comp. bio)

            cores (int) - number of cores per batch

            memory (int) - memory per batch, GB

            kwargs: keyword arguments used at runtime

        """

        # define paths
        path = abspath(path)
        run_path = path.rsplit('/', maxsplit=1)[0]
        job_script_path = join(path, 'scripts', 'submit.sh')

        # copy run script to scripts directory
        scripts_path = abspath(__file__).rsplit('/', maxsplit=2)[0]
        run_script = join(scripts_path, 'scripts', script_name)
        shutil.copy(run_script, join(path, 'scripts'))

        # determine queue
        if walltime <= 4:
            queue = 'short'
        elif walltime <= 48:
            queue = 'normal'
        else:
            queue = 'long'

        # declare outer script that reads PATH from file
        job_script = open(job_script_path, 'w')
        job_script.write('#!/bin/bash\n')

        # move to benchmarking directory
        job_script.write('cd {:s} \n\n'.format(run_path))

        # begin outer script for processing job
        job_script.write('while IFS=$\'\\t\' read P\n')
        job_script.write('do\n')
        job_script.write('b_id=$(echo $(basename ${P}) | cut -f 1 -d \'.\')\n')
        job_script.write('   JOB=`msub - << EOJ\n\n')

        # =========== begin submission script for individual job ==============
        job_script.write('#! /bin/bash\n')
        job_script.write('#MSUB -A {:s} \n'.format(allocation))
        job_script.write('#MSUB -q {:s} \n'.format(queue))
        job_script.write('#MSUB -l walltime={0:02d}:00:00 \n'.format(walltime))
        job_script.write('#MSUB -m abe \n')
        #job_script.write('#MSUB -M sebastian@u.northwestern.edu \n')
        job_script.write('#MSUB -o ./benchmark/log/${b_id}/outlog \n')
        job_script.write('#MSUB -e ./benchmark/log/${b_id}/errlog \n')
        job_script.write('#MSUB -N ${b_id} \n')
        job_script.write('#MSUB -l nodes=1:ppn={:d} \n'.format(cores))
        job_script.write('#MSUB -l mem={:d}gb \n\n'.format(memory))

        # load python module and clones virtual environment
        job_script.write('module load python/anaconda3.6\n')
        job_script.write('source activate ~/pythonenvs/clones_env\n\n')

        # move to job directory
        job_script.write('cd {:s} \n\n'.format(run_path))

        # run script
        cmd = 'python ./benchmark/scripts/{:s}'.format(script_name)+' ${P}'
        for k, v in kwargs.items():
            cmd += ' -{} {}'.format(k, v)
        job_script.write(cmd+' \n')

        job_script.write('EOJ\n')
        job_script.write('`\n\n')
        # ============= end submission script for individual job --============

        # print job id
        job_script.write('done < ./benchmark/jobs/index.txt \n')
        job_script.write('echo "All jobs submitted as of `date`"\n')
        job_script.write('exit\n')

        # close the file
        job_script.close()

        # change the permissions
        chmod(job_script_path, 0o755)

    def build_batches(self):
        """ Writes benchmark paths for each batch. """

        # get directories for all batches and logs
        batches_dir = join(self.path, 'batches')
        jobs_dir = join(self.path, 'jobs')
        logs_dir = join(self.path, 'log')

        # create index file for batches
        index_path = join(jobs_dir, 'index.txt')
        index = open(index_path, 'w')

        # write file containing benchmark paths for each batch
        for culture_id, batch in enumerate(self.batches.ravel()):

            # write batch benchmark paths to job file
            for ambiguity_id in range(self.num_ambiguities):

                # determine batch id
                batch_id = (culture_id * self.batches.size) + ambiguity_id

                # append job file to index
                job_path = join(jobs_dir, '{:d}.txt'.format(batch_id))
                index.write('{:s}\n'.format(relpath(job_path, self.sweep_path)))

                # open job file
                job_file = open(job_path, 'w')

                # write batch paths to job file
                benchmark_path = self.benchmark_paths[(culture_id, ambiguity_id)]
                job_file.write('{:s}\n'.format(benchmark_path))

                # create log directory for job
                mkdir(join(logs_dir, '{:d}'.format(batch_id)))

                # close job file
                job_file.close()
                chmod(job_path, 0o755)

        # close index file
        index.close()
        chmod(index_path, 0o755)

    def make_subdirectory(self):
        """ Create job subdirectory. """

        # create directory (overwrite existing one)
        path = join(self.sweep_path, 'benchmark')
        if isdir(path):
            shutil.rmtree(path)
        mkdir(path)
        self.path = path

        # make subdirectories for simulations and scripts
        mkdir(join(path, 'scripts'))
        mkdir(join(path, 'batches'))
        mkdir(join(path, 'jobs'))
        mkdir(join(path, 'log'))

    def build(self,
              walltime=10,
              allocation='p30653',
              cores=1,
              memory=4,
              **kwargs):
        """
        Build job directory tree. Instantiates and saves a simulation instance for each parameter set, then generates a single shell script to submit each simulation as a separate job.

        Args:

            walltime (int) - estimated job run time

            allocation (str) - project allocation

            cores (int) - number of cores per batch

            memory (int) - memory per batch, GB

            kwargs (dict) - keyword arguments for benchmarking

        """

        # create benchmarking subdirectory
        self.make_subdirectory()

        # store parameters
        self.kwargs = kwargs

        # build batch benchmarks
        for batch_id, batch in enumerate(self.batches.ravel()):

            # make benchmarks for each ambiguity value in the batch
            for ambiguity_id, ambiguity in enumerate(self.ambiguities):

                # determine job index
                job_id = (batch_id * self.batches.size) + ambiguity_id

                # make job directory
                job_path = join(self.path, 'batches', '{:d}'.format(job_id))
                mkdir(job_path)

                # build benchmark for current ambiguity
                args = (job_path, batch, ambiguity, self.num_replicates)
                self.build_job(*args, **kwargs)

                # store job path
                self.benchmark_paths[(batch_id, ambiguity_id)] = relpath(job_path, self.sweep_path)

        # save serialized job
        self.save(join(self.path, 'benchmark_job.pkl'))

        # build parameter file for each batch
        self.build_batches()

        # build job run script
        self.build_run_script(self.path, self.script_name)

        # build custom template script
        self.build_run_script(self.path, 'template.py', 'run_template.sh')

        # build job submission script
        self.build_submission_script(self.path,
                                     self.script_name,
                                     walltime=walltime,
                                     allocation=allocation,
                                     cores=cores,
                                     memory=memory)

    @classmethod
    def build_job(cls, dirpath, batch, ambiguity, num_replicates, **kwargs):
        """
        Builds and saves a BatchBenchmark instance for a given batch and fluorescence ambiguity coefficient.

        Args:

            dirpath (str) - path to job directory

            batch (growth.sweep.Batch) - batch of growth replicates

            ambiguity (float) - fluorescence ambiguity coefficient

            num_replicates (int) - number of fluorescence replicates

            kwargs: keyword arguments for BatchBenchmark

        """
        benchmark = BatchBenchmark(batch, ambiguity, num_replicates, **kwargs)
        benchmark.save(dirpath)

    def load_job(self, batch_id, ambiguity_id):
        """
        Load benchmarking job from file.

        Args:

            batch_id (int) - batch index

            ambiguity_id (int) - ambiguity coefficient index

        Returns:

            benchmark (BatchBenchmark)

        """
        job_key = (batch_id, ambiguity_id)
        path = join(self.sweep_path, self.benchmark_paths[job_key])
        benchmark = BatchBenchmark.load(path)
        benchmark.batch.root = self.sweep_path
        return benchmark

    def aggregate(self):
        """ Aggregate results from all batches. """

        nrows, ncols = self.batches.shape

        failures = open(join(self.path, 'failed.txt'), 'w')

        # compile results from all batches
        data = []
        for row_id in range(nrows):
            for column_id in range(ncols):
                batch_id = row_id*ncols + column_id

                for ambiguity_id in range(self.num_ambiguities):

                    # load benchmark for current batch
                    batch_benchmark = self.load_job(batch_id, ambiguity_id)

                    if batch_benchmark.results is None:
                        path = self.benchmark_paths[(batch_id, ambiguity_id)]
                        failures.write('{:s}\n'.format(path))

                    else:
                        # append results to list
                        batch_data = batch_benchmark.results
                        batch_data['batch_id'] = batch_id
                        batch_data['row_id'] = row_id
                        batch_data['column_id'] = column_id
                        batch_data['ambiguity_id'] = ambiguity_id
                        data.append(batch_data)

        failures.close()

        data = pd.concat(data).reset_index()

        # save results to file
        self._results = data
        self._results.to_hdf(join(self.path, 'data.hdf'), key='benchmark')

    @property
    def results(self):
        """ Returns benchmarking results object. """

        # make sure results have been compiled
        if self._results is None:
            flag = str2bool(input('Results not compiled. Compile now?'))
            if flag:
                self.aggregate()
            else:
                return None

        return BenchmarkingResults(self._results,
                                   self.batches.shape,
                                   clone_sizes=self.mean_clone_sizes,
                                   ambiguities=self.ambiguities)



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.colors import Normalize


class SweepVisualization:
    """ Methods for visualizing fluorescence values. """

    @property
    def max_fluroescence(self):
        """ Peak fluorescence across sweep. """
        batch = self.load_benchmark(0, self.num_ambiguities-1)
        return np.percentile(batch.classifier.values, 99)

    @property
    def fnorm(self):
        """ Fluorescence normalization across sweep. """
        return Normalize(vmin=0, vmax=self.max_fluroescence)

    def plot_fluorescence(self,
                          replicate_id=0,
                          resolution=4,
                          norm=None,
                          cmap=plt.cm.viridis,
                          figsize=(10, 10),
                          **kwargs):

        # determine shape
        shape = self.ambiguities.size, self.batches.shape[1]
        xs = range(0, shape[1], resolution)
        ys = range(0, shape[0], resolution)
        nrows, ncols = len(ys), len(xs)

        # determine normalization
        if norm == 'global':
            norm = self.fnorm

        # create figure
        fig = plt.figure(figsize=figsize)
        gs = GridSpec(nrows, ncols, wspace=0, hspace=0)

        # plot fluorescence values
        for i, ambiguity_id in enumerate(xs):
            for j, batch_id in enumerate(ys[::-1]):

                # add axis
                ax = fig.add_subplot(gs[i, j])

                # load BatchBenchmark then generate and plot measurement data
                batch = self.load_benchmark(batch_id, ambiguity_id)
                simulation = batch.benchmark_simulation(replicate_id)
                simulation.plot_measurements(ax=ax, norm=norm, **kwargs)

                # format axis
                ax.set_xlim(-1.1, 1.1)
                ax.set_ylim(-1.1, 1.1)
                ax.axis('off')

        return fig



import numpy as np


class AttributeLabeler:
    """
    Assigns label to cell measurement data based on an existing attribute.

    Attributes:

        label (str) - name of label field to be added

        attribute (str) - existing cell attribute used to determine labels

        labeler (vectorized func) - callable that maps attribute values to labels

    """

    def __init__(self, label, attribute, labels):
        """
        Instantiate labeler.

        Args:

            label (str) - name of label attribute to be added

            attribute (str) - existing cell attribute used to determine labels

            labels (dict) - {attribute value: label value} pairs

        """

        # store label and attribute field names
        self.label = label
        self.attribute = attribute

        # vectorize labeling function
        self.labeler = np.vectorize(labels.get)

    def __call__(self, data):
        """
        Assign labels by adding <label> field to cell measurement data.

        Args:

           data (pd.DataFrame) - cells measurement data with <attribute> field

        """
        return self.assign_labels(data)

    def assign_labels(self, data):
        """
        Assign labels by adding <label> field to cell measurement data.

        Args:

            data (pd.DataFrame) - cells measurement data with <attribute> field

        """
        data[self.label] = self.labeler(data[self.attribute])


class CelltypeLabeler(AttributeLabeler):
    """
    Assigns <celltype> to cell measurement data based on <genotype> attribute.

    Attributes:

        label (str) - name of label field to be added

        attribute (str) - existing cell attribute used to determine labels

        labeler (vectorized func) - callable that maps attribute values to labels

    """

    def __init__(self, label='celltype', attribute='genotype', labels=None):
        """
        Instantiate celltype labeler.

        Args:

            label (str) - name of label attribute to be added

            attribute (str) - existing cell attribute used to determine labels

            labels (dict) - {genotype value: label} pairs

        """

        # use default genotype labels
        if labels is None:
            labels = {0:'m', 1:'h', 2:'w', -1:'none'}

        super().__init__(label, attribute, labels)



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec

from ...visualization import *


def figure(func):
    """ Decorator for creating axis. """
    def wrapper(*args, ax=None, **kwargs):
        if ax is None:
            fig, ax = plt.subplots(figsize=(2., 1.25))
        func(*args, ax=ax, **kwargs)
    return wrapper

def surface_figure(func):
    """ Decorator for creating joint axis. """
    def wrapper(*args, ax=None, **kwargs):
        if ax is None:
            fig, ax = plt.subplots(figsize=(2., 2.))
        func(*args, ax=ax, **kwargs)
    return wrapper


class MixtureVisualization:
    """ Visualization methods for mixture models. """

    @property
    def summary(self):
        """ Returns text-based summary of mixture model. """
        m = ' :: '.join(['{:0.2f}'.format(x) for x in self.means])
        s = ' :: '.join(['{:0.2f}'.format(np.sqrt(x)) for x in self.stds])
        w = ' :: '.join(['{:0.2f}'.format(x) for x in self.weights_])
        summary = 'Means: {:s}'.format(m)
        summary += '\nStds: {:s}'.format(s)
        summary += '\nWeights: {:s}'.format(w)
        summary += '\nlnL: {:0.2f}'.format(self.log_likelihood)
        return summary

    @figure
    def plot_component_pdf(self, idx,
                           weighted=True,
                           log=True,
                           ax=None,
                           **kwargs):
        """ Plots PDF for specified component. """

        # retrieve pdf for specified component
        pdf = self.get_component_pdf(idx, weighted=weighted)

        # plot component pdf
        if log:
            ax.plot(self.support, pdf, **kwargs)
        else:
            ax.plot(self.scale_factor, pdf/self.scale_factor, **kwargs)

        self.format_ax(ax, log=log)

    @figure
    def plot_pdf(self, log=True, ax=None, **kwargs):
        """ Plots overall PDF for mixture model. """

        if log:
            ax.plot(self.support, self.pdf, **kwargs)
        else:
            ax.plot(self.scale_factor, self.pdf/self.scale_factor, **kwargs)

        self.format_ax(ax, log=log)

    @figure
    def plot(self, log=True, ax=None,
             pdf_color='k', component_color='r', **kwargs):
        """ Plots PDF for mixture model as well as each weighted component. """
        self.plot_pdf(log=log, ax=ax, color=pdf_color)
        for i in range(self.n_components):
            self.plot_component_pdf(i, log=log, ax=ax, color=component_color)

    @figure
    def plot_data(self, log=True, ax=None, **kwargs):
        """ Plot binned values. """

        if log:
            data = self.values
        else:
            data = np.exp(self.values)

        bins = np.linspace(self.support.min(), self.support.max(), num=50)
        _ = ax.hist(data, bins=bins, density=True, **kwargs)

    def format_ax(self, ax, log=True):
        if log:
            ax.set_xlim(self.support.min(), self.support.max())
        else:
            ax.set_xlim(0, self.scale_factor.max())
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)


class BivariateVisualization:
    """ Visualization methods for bivariate mixture models. """

    @surface_figure
    def plot_pdf_surface(self, ax=None, contours=None, **kwargs):

        ax.imshow(self.pdf, extent=self.extent, **kwargs)

        if contours is not None:
            ax.contour(self.pdf, contours, extent=self.extent[[0,1,3,2]], colors='r')

        self.format_joint_ax(ax)

    @figure
    def plot_margin(self, margin=0,
                    invert=False,
                    log=True,
                    pdf_color='k',
                    pdf_linestyle='--',
                    pdf_lw=0.5,
                    component_color='r',
                    component_lw=0.5,
                    min_density=0.01,
                    ax=None):

        # extract x/y margins
        if margin == 0:
            support, pdf = self.get_xmargin(log=log)
        else:
            support, pdf = self.get_ymargin(log=log)

        # define colors for component lines
        if type(component_color) == str:
            component_color *= self.n_components

        # define parameters
        above = (pdf >= min_density)
        pdf_kw = dict(color=pdf_color, lw=pdf_lw, linestyle=pdf_linestyle)
        comp_kw = dict(lw=component_lw)

        # invert x/y axes (for vertical margin)
        if invert:
            ax.plot(pdf[above], support[above], '-', **pdf_kw)
            for i in range(self.n_components):
                mpdf = self.get_component_marginal_pdf(i, margin, True)
                ind = (mpdf >= min_density)
                ax.plot(mpdf[ind], support[ind], color=component_color[i], **comp_kw)
        else:
            ax.plot(support[above], pdf[above], '-', **pdf_kw)
            for i in range(self.n_components):
                mpdf = self.get_component_marginal_pdf(i, margin, True)
                ind = (mpdf >= min_density)
                ax.plot(support[ind], mpdf[ind], color=component_color[i], **comp_kw)

        return ax

    def visualize(self,
                  size_ratio=4,
                  figsize=(2, 2),
                  contours=None,
                  **kwargs):
        """ Visualize joint and marginal distributions. """

        # create figure
        fig = plt.figure(figsize=figsize)
        ratios = [size_ratio/(1+size_ratio), 1/(1+size_ratio)]
        gs = GridSpec(2, 2, width_ratios=ratios, height_ratios=ratios[::-1], wspace=0, hspace=0)
        ax_joint = fig.add_subplot(gs[1, 0])
        ax_xmargin = fig.add_subplot(gs[0, 0])
        ax_ymargin = fig.add_subplot(gs[1, 1])
        ax_xmargin.axis('off')
        ax_ymargin.axis('off')

        # plot multivariate pdf surface
        self.plot_pdf_surface(ax=ax_joint, contours=contours, **kwargs)

        # plot marginal pdfs
        self.plot_margin(0, ax=ax_xmargin)
        self.plot_margin(1, invert=True, ax=ax_ymargin)

        ax_xmargin.set_xlim(self.lbound, self.ubound)
        ax_ymargin.set_ylim(self.lbound, self.ubound)

        return fig

    @figure
    def plot_data(self, ax=None, **kwargs):
        """ Scatter datapoints on <ax>. """
        ax.scatter(*self.values.T, **kwargs)

    def format_joint_ax(self, ax):
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.invert_yaxis()
        ax.set_xlabel('ln X')
        ax.set_ylabel('<ln X> among neighbors')

        ax.set_xticks(self.tick_positions)
        ax.set_yticks(self.tick_positions)
        ax.spines['left'].set_position(('outward', 2))
        ax.spines['bottom'].set_position(('outward', 2))

    def format_marginal_ax(self, ax):
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.set_xlabel('ln X')

    @property
    def tick_positions(self):
        """ Tick positions. """
        lbound = np.ceil(self.supportx.min())
        ubound = np.floor(self.supportx.max())
        step = (ubound - lbound) // 4
        return np.arange(lbound, ubound+1, step)



from os.path import join, exists
from os import mkdir
import numpy as np

from ...utilities import IO

from ..classification import UnivariateMixtureClassifier
from .visualization import ModelSelectionVisualization


class SelectionIO:
    """
    Methods for saving and loading a model selection instance.
    """
    def save(self, dirpath, image=False, **kwargs):
        """
        Save classifier to specified path.

        Args:

            dirpath (str) - directory in which classifier is to be saved

            image (bool) - if True, save model image

            kwargs: keyword arguments for image rendering

        Returns:

            path (str) - model selection directory

        """

        # create directory for model selection
        path = join(dirpath, 'models')
        if not exists(path):
            mkdir(path)

        # save parameters
        io = IO()
        io.write_json(join(path, 'parameters.json'), self.parameters)

        # save values
        np.save(join(path, 'values.npy'), self.values)

        # save models
        for n, model in self._models.items():
            model.save(path, image=image, extension=n, **kwargs)

        return path

    @staticmethod
    def load_model(path):
        """ Load model from <path> directory. """
        return UnivariateMixtureClassifier.load(path)

    @classmethod
    def load(cls, path):
        """
        Load model selection instance from file.

        Args:

            path (str) - model selection directory

        Returns:

            selector (UnivariateModelSelection derivative)

        """
        io = IO()

        # load values and parameters
        values = io.read_npy(join(path, 'values.npy'))
        parameters = io.read_json(join(path, 'parameters.json'))
        attribute = parameters.pop('attribute')

        # load models
        n_min = parameters['min_num_components']
        n_max = parameters['max_num_components']

        models = {}
        for num_components in range(n_min, n_max+1):
            model_path = join(path, 'classifier_{:d}'.format(num_components))
            if exists(model_path):
                model = cls.load_model(model_path)
                model._values = values
                model.model.values = np.log(values)
                models[num_components] = model

        return cls(values, attribute, models=models, **parameters)


class UnivariateModelSelection(SelectionIO, ModelSelectionVisualization):
    """
    Class for performing univariate mixture model selection. The optimal model is chosen based on BIC score.
    """

    def __init__(self, values, attribute,
                 min_num_components=3,
                 max_num_components=8,
                 num_labels=3,
                 models=None):
        """
        Perform model selection by choosing the model that minimizes BIC score.

        Args:

            values (np.ndarray[float]) - 1D array of sample values

            attribute (str) - attribute label for sample values

            min_num_components (int) - minimum number of components in mixture

            max_num_components (int) - maximum number of components in mixture

            num_labels (int) - maximum number of unique labels to be assigned

            models (dict) - pre-fitted Classification instances keyed by number of components

        """

        self.values = values
        self.attribute = attribute
        self.min_num_components = min_num_components
        self.max_num_components = max_num_components
        self.num_labels = num_labels
        self.num_components = range(min_num_components, max_num_components+1)

        # fit models
        if models is None:
            models = self.fit_models()
        self._models = models

    @staticmethod
    def fit_model(values, num_components, num_labels, **kwargs):
        """ Fit model with specified number of components. """
        return UnivariateMixtureClassifier(values,
                                    num_components=num_components,
                                    num_labels=num_labels,
                                    **kwargs)

    def fit_models(self):
        """ Fit model with each number of components. """

        # define parameters
        args = (self.values,)
        kwargs = dict(attribute=self.attribute)

        # fit models
        models_dict = {}
        for num_components in self.num_components:
            model = self.fit_model(self.values,
                                   num_components,
                                   self.num_labels, **kwargs)
            models_dict[num_components] = model

        return models_dict

    @property
    def parameters(self):
        """ Dictionary of instance parameters. """
        return {
            'attribute': self.attribute,
            'min_num_components': self.min_num_components,
            'max_num_components': self.max_num_components,
            'num_labels': self.num_labels}

    @property
    def models(self):
        """ List of models ordered by number of components. """
        return [m for n, m in sorted(self._models.items())]

    @property
    def BIC(self):
        """ BIC scores for each model. """
        return np.array([model.model.BIC for model in self.models])

    @property
    def BIC_optimal(self):
        """ Model with BIC optimal number of components. """
        return self.models[np.argmin(self.BIC)]

    @property
    def AIC(self):
        """ AIC scores for each model. """
        return np.array([model.model.AIC for model in self.models])

    @property
    def AIC_optimal(self):
        """ Model with AIC optimal number of components. """
        return self.models[np.argmin(self.AIC)]



from os.path import join, exists
from os import mkdir
from shutil import rmtree
import numpy as np
import pandas as pd
from scipy.spatial import qhull
from skimage.measure import regionprops
import PIL

from .silhouette_read import ReadSilhouetteData

from ..utilities import IO
from ..validation.arguments import str2bool


class WriteSilhouette:
    """
    Methods for writing a stack to Silhouette readable format.

    The Silhouette container includes a FEED file:

    FEED.json ::

        "orientation": {"flip_about_xy": false, "flip_about_yz": false},
        "layer_ids": [ 0,1,2... ],
        "params": { param_name: param_value ... } }

    """

    @property
    def silhouette_path(self):
        """ Path to Silhouette directory. """
        return join(self.path, '{:s}.silhouette'.format(self._id))

    def load_silhouette_labels(self):
        """
        Load manually assigned labels from file.

        Returns:

            labels (pd.Series) - labels keyed by (layer_id, segment_id)

        """

        # load silhouette data
        silhouette = ReadSilhouetteData(self.silhouette_path, recompile=True)

        # convert labels to numeric scale
        label_to_value = dict(M=0, H=1, W=2)
        labels = silhouette.labels.apply(label_to_value.get)
        labels.name = 'manual_label'

        return labels

    @property
    def _feed(self):
        return {
            "layer_ids": list(range(len(self.included))),
            "orientation": {
                "flip_about_xy": False,
                "flip_about_yz": False},
            "params": {
                "cell_area": 200,
                "cut_line_thickness": 2,
                "max_cut_distance": 150,
                "max_loop": 20,
                "meanshift_sp": 5,
                "meanshift_sr": 30,
                "min_hull_distance": 5,
                "min_intensity": 30,
                "opening_size": 3,
                "small_cell_area": 25,
                "subsegment_cell_area": 400,
                "subsegment_max_deep": 3,
                "subsegment_opening_size": 0,
                "total_percentage": 0}}
    
    def build_feud(self, label=None):
        """ Compile feud file with <label> field serving as annotations. """
    
        # initialize feud file
        feud = {"filter":{"r":0,"g":0,"b":0}, "layers": []}
        
        # compile labels from each layer
        for layer_id, layer in enumerate(self):
            
            # compile layer's contour labels
            if label is None or label not in layer.data.columns:            
                contours = []
            else:        
                contours = [{"id": int(x), "label": str(y)} for x,y in layer.data[['segment_id', label]].values]        
            
            # add to feud
            feud["layers"].append({"id": layer_id, "contours": contours})
            
        return feud

    def write_silhouette(self,
                         dst=None,
                         label=None,
                         include_image=True,
                         channel_dict=None):
        """
        Write silhouette file.

        Args:

            dst (str) - destination directory

            label (str) - field containing cell type annotations

            include_image (bool) - save RGB image of each layer

            channel_dict (dict) - RGB channel names, keyed by channel index. If none provided, defaults to the first three channels in RGB order.

        """

        # if no directory is specified, use stack directory
        if dst is None:
            dst = self.silhouette_path

        # create silhouette directory
        if exists(dst):
            overwrite = str2bool(input('Destination exists. Overwrite? '))
            if overwrite:
                rmtree(dst)
            else:
                return
        mkdir(dst)

        # make sure image is loaded
        if self.stack is None:
            self.load_image()

        # write feed file
        io = IO()
        io.write_json(join(dst, 'feed.json'), self._feed)

        # write feud file
        feud = self.build_feud(label=label)
        io.write_json(join(dst, 'feud.json'), feud)

        # write layer files
        for layer_id, layer in enumerate(self):
            layer.write_silhouette(dst,
                layer_id=layer_id,
                include_image=include_image,
                channel_dict=channel_dict)


class WriteSilhouetteLayer:
    """
    Methods for writing a Layer to Silhouette readable format. A layer file is structured as follows:

    LAYER_ID.json :

        {
        "id": LAYER_ID
        "imageFilename": "LAYER_ID.png"
        "contours": [ ... contours ... ]

            {"centroid": [CONTOUR_CENTROID_X, CONTOUR_CENTROID_Y],
            "color_avg": {"b": X, "g": X, "r": X},
            "color_std": {"b": X, "g": X, "r": X},
            "id": CONTOUR_ID,
            "pixel_count": CONTOUR_AREA,
            "points": [[x1, y1], [x2, y2] ... ]}}


    """

    @staticmethod
    def _to_key(channel):
        """ Returns DataFrame key for <channel>. """
        return 'ch{:d}'.format(channel)

    @staticmethod
    def _construct_hull(properties):
        """ Returns points on convex hull. """
        hull = qhull.ConvexHull(properties.coords)
        return hull.points[hull.vertices].astype(int).tolist()

    def build_contours(self, channel_dict):
        """
        Convert dataframe to a list of contours (Silhouette format).

        Args:

            channel_dict (dict) - RGB channel names, keyed by channel index

        Returns:

            contours (list) - list of contour dictionaries

        """

        # compile contour data
        assert self.labels is not None, 'Labels are not defined.'
        properties = regionprops(self.labels.T)
        ctr_data = pd.Series({p.label: self._construct_hull(p) \
                             for p in properties}, name='points')
        data = self.data.join(ctr_data, on='segment_id')

        # coerce into Silhouette contour-list format
        data['id'] = data.segment_id
        data['centroid'] = data[['centroid_x', 'centroid_y']].values.tolist()

        # identify key maps for RGB channels
        mean_dict = {self._to_key(k): v for k, v in channel_dict.items()}
        std_dict = {k+'_std': v+'_std' for k, v in mean_dict.items()}
        keys = ['ch{:d}'.format(x) for x in range(self.color_depth)] + list(mean_dict.values())

        # append RGB mean intensities to dataframe
        data[list(mean_dict.values())] = data[list(mean_dict.keys())]
        data[list(std_dict.values())] = data[list(std_dict.keys())]
        data['color_avg'] = data[keys].to_dict(orient='records')

        # append RGB intensity variation to dataframe
        std_data = data[[k+'_std' for k in keys]]
        std_data.columns = keys
        data['color_std'] = std_data.to_dict(orient='records')

        # compile contour list
        keys = ['id', 'centroid', 'pixel_count', 'points',
                'color_avg', 'color_std']
        contours = data[keys].to_dict(orient='records')

        return contours

    def _write_png(self, dst, channel_dict):
        """
        Write layer image to Silhouette-compatible RGB image in PNG format.

        Args:

            dst (str) - png filepath

            channel_dict (dict) - RGB channel names, keyed by channel index

        """
        rev_channel_dict = {v: k for k, v in channel_dict.items()}
        image_8bit = np.round(self.im * 2**8).astype(np.uint8)
        image_RGB = image_8bit[..., [rev_channel_dict[c] for c in 'rgb']]
        image = PIL.Image.fromarray(image_RGB)
        image.save(dst, format='png')

    def write_silhouette(self, dst,
                         layer_id=None,
                         include_image=True,
                         channel_dict=None,
                         ):
        """
        Write silhouette compatible JSON to target directory.

        Args:

            dst (str) - destination directory

            layer_id (int) - ID optionally used to override true layer ID

            include_image (bool) - save layer image as png

            channel_dict (dict) - RGB channel names, keyed by channel index. If none provided, defaults to the first three channels in RGB order.

        """

        # define layer ID (overrides true stack layer)
        if layer_id is None:
            layer_id = self._id

        # define rgb map
        if channel_dict is None:
            channel_dict = dict(enumerate('rgb'))

        # define paths
        filepath = join(dst, '{:d}.json'.format(layer_id))
        
        # write contour data
        io = IO()
        contours = {
            'id': layer_id,
            'imageFilename': '{:d}.png'.format(layer_id),
            'contours': self.build_contours(channel_dict)}
        io.write_json(filepath, contours)

        # write image
        if include_image:
            img_path = join(dst, '{:d}.png'.format(layer_id))
            self._write_png(img_path, channel_dict=channel_dict)



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize
from skimage.transform import resize


class Mask:
    """
    Image mask.
    """

    def __init__(self, mask):
        self.mask = mask

    def add_contour(self, ax, lw=1, color='r'):
        ax.contour(self.mask, [0.5], linewidths=[lw], colors=[color])

    def add_contourf(self, ax, alpha=0.5, colors='none', hatches=['//']):
        ax.contourf(self.mask, [-.5, .5], colors=['w'], hatches=hatches, alpha=alpha)


class FunctionMask:
    """
    Mask constructed by thresholding the output of a function.
    """

    def __init__(self, shape, func, res=100):
        self.res = res
        self.shape = shape
        self.values = self.evaluate(func)
        self.vmin, self.vmax = self.values.min(), self.values.max()
        self.apply_threshold(func.threshold)

    def evaluate(self, func):
        yy, xx = np.meshgrid(*(np.arange(0, s, self.res) for s in self.shape))
        grid = np.vstack((xx.ravel(), yy.ravel())).T
        values = resize(func(grid).reshape(yy.shape), self.shape)
        return np.ma.array(values).T

    def apply_threshold(self, threshold):
        self.values.mask = (self.values < threshold)

    def add_contour(self, ax, lw=1, color='r'):
        ax.contour(self.values.mask, [0.5], linewidths=[lw], colors=[color])

    def add_contourf(self, ax, alpha=0.5, colors=['w'], hatches=['//']):
        ax.contourf(~self.values.mask, [-.5, .5], colors=colors, hatches=hatches, alpha=alpha)

    def plot_density(self, mask_alpha=0.5, cmap=plt.cm.plasma, figsize=(5, 5)):
        fig, ax = plt.subplots(figsize=figsize)
        ax.imshow(self.values.data, cmap=cmap, vmin=self.vmin, vmax=self.vmax)
        return ax


class ForegroundMask(FunctionMask):
    pass


class RBFMask:
    """
    Mask based on thresholding an RBF fit.
    """

    def __init__(self, shape, rbf, mask=None, res=100):
        self.rbf = rbf
        self.res = res
        self.shape = shape
        self.zi = self.evaluate(rbf)
        self.label()

        if mask is not None:
            self.apply_mask(mask)

    def evaluate(self, func):
        yy, xx = np.meshgrid(*(np.arange(0, s, self.res) for s in self.shape))
        zi = func(xx, yy)
        return resize(zi, self.shape).T

    def label(self):
        labels = np.searchsorted(self.rbf.thresholds, self.zi)
        self.labels = np.ma.array(labels)

    def apply_mask(self, mask):
        self.labels.mask = mask

    def add_contour(self, ax, lw=3, alpha=1, colors=None, cmap=None):
        if cmap is None:
            cmap = plt.cm.Reds

        lmax = self.labels.max()
        levels = np.arange(-0.5, lmax+1, 1)
        if colors is None:
            norm = Normalize(levels.min(), levels.max())
            colors = [cmap(norm(l)) for l in levels]
        ax.contour(self.labels, levels, colors=colors, alpha=alpha, linewidths=lw)

    def add_contourf(self, ax, alpha=0.5, colors=None, cmap=None):

        if cmap is None:
            cmap = plt.cm.Reds

        lmax = self.labels.max()
        levels = np.arange(-0.5, lmax+1, 1)
        if colors is None:
            norm = Normalize(levels.min(), levels.max())
            colors = [cmap(norm(l)) for l in levels]
        ax.contourf(self.labels, levels, colors=colors, alpha=alpha)

    def plot_density(self, cmap=None, figsize=(5, 5)):
        if cmap is None:
            cmap = plt.cm.Reds

        fig, ax = plt.subplots(figsize=figsize)
        ax.imshow(self.zi, cmap=cmap)
        ax.axis('off')
        return ax

    def plot_regions(self, s=10, lw=1, cmap=None, figsize=(5, 5)):
        if cmap is None:
            cmap = plt.cm.Reds
        fig, ax = plt.subplots(figsize=figsize)
        ax.imshow(self.labels, cmap=cmap, vmin=0, vmax=self.labels.max())
        self.rbf.plot_points(ax, lw=lw, s=s, cmap=cmap)
        ax.axis('off')
        return ax



import numpy as np
import pandas as pd
from sklearn.metrics import f1_score

from ..visualization import *
from ..visualization.figures import Figure


class Scoring:
    """
    Object for scoring classification accurancy.

    Attributes:

        data (pd.DataFrame) - data with 'measured' and 'predicted' attributes

        n (int) - number of classifications

        percent_correct (float) - frequency of correct classification

        matrix (ErrorMatrix) - classification adjacency matrix

    """

    def __init__(self, measured, predicted):
        """
        Instantiate scoring prediction.

        Args:

            measured (array like) - manually assigned class labels

            predicted (array like) - predicted class labels

        """
        columns = ('measured', 'predicted')

        if type(measured) in (np.ndarray, list, tuple):
            data = np.vstack((measured, predicted)).T
            data = pd.DataFrame(data, columns=columns)
        elif type(measured) == pd.Series:
            data = pd.DataFrame([measured, predicted], index=columns).T
        else:
            raise ValueError('Type of <measured> not recognized.')

        self.data = data
        self.n = len(self.data)
        self.compare()

    def __add__(self, x):
        """ Combine additively with another scoring matrix. """
        merged = pd.concat([self.data, x.data])
        return Scoring(merged[['measured', 'predicted']])

    @property
    def error(self):
        """ Normalized total classification error. """
        return self.data.difference.sum() / (2*self.n)

    @property
    def percent_correct(self):
        """ Fraction of measurements with correct annotation. """
        return self.data.correct.sum()/self.n

    @property
    def f1(self):
        """ F1 score. """
        values = self.data[['measured', 'predicted']].values
        return f1_score(*values.T, average='weighted')

    @property
    def MAE(self):
        """ Mean absolute error averaged to correct for imbalance. """
        return self.data.groupby('measured')['difference'].mean().mean() / 2

    @property
    def percent_correct_per_layer(self):
        """ Percent correct per layer. """
        return (self.data.correct).groupby(['stack', 'layer']).mean()

    @property
    def MAE_per_layer(self):
        """ Mean absolute error per layer. """
        MAE = lambda x: x.groupby('measured')['difference'].mean().mean() / 2
        return self.data.groupby(['stack', 'layer']).apply(MAE)

    def compare(self):
        """ Evaluate frequency of correct classification. """
        self.data['difference'] = abs(self.data.measured-self.data.predicted)
        self.data['correct'] = (self.data.difference==0)

    def plot_matrix(self, **kwargs):
        """
        Plot classification adjacency matrix.

        kwargs: keyword arguments for adjacency matrix
        """
        measured = self.data.measured.values.astype(int)
        predicted = self.data.predicted.values.astype(int)
        self.matrix = AdjacencyMatrix(measured, predicted, **kwargs)

    @default_figure
    def plot_label_distribution(self,
                                mcolor='tomato',
                                pcolor='darkmagenta',
                                width=0.4,
                                legend=False,
                                ax=None):
        """
        Plot distribution of labels among annotated cells.

        Args:

            mcolor (str) - color for measured labels

            pcolor (str) - color for predicted labels

            width (float) - bar width

            legend (bool) - if True, include legend

        """

        # group data by label-type and obtain counts
        m = self.data.measured.groupby(self.data.measured).count().values
        p = self.data.predicted.groupby(self.data.predicted).count().values

        # plot bars
        ind = np.arange(m.size)
        _ = ax.bar(ind-width/2, m, width, color=mcolor, label='Measured')
        _ = ax.bar(ind+width/2, p, width, color=pcolor, label='Predicted')

        if legend:
            ax.legend(loc=2, frameon=False, fontsize=6,
                       bbox_to_anchor=(0.05, 1.1, 1., .102),
                       ncol=1, borderaxespad=0.)

        # format axis
        ax.set_xticks(ind)
        ax.set_xticklabels(['{:d}x'.format(i) for i in ind])
        ax.set_ylabel('Number of labels')


class AdjacencyMatrix(Figure):
    """
    Classification adjacency matrix. Generates a figure showing the frequency of overlap between predicted and manually assigned labels.

    Attributes:

        counts (np.ndarray[int]) - pairwise overlap between manually assigned and predicted class labels

        fig (matplotlib.figures.Figure) - adjacency matrix figure

    """

    def __init__(self, measured, predicted,
                 ax=None,
                 text=None,
                 figsize=(1.5, 1.5),
                 **kwargs):
        """
        Instantiate error matrix.

        Args:

            measured (array like) - measured class labels

            predicted (array like) - predicted class labels

            ax (matplotlib.axes.AxesSubplot) - if None, create figure

            text (str) - indicates whether 'counts' or 'rates' labels are added

            figsize (tuple) - figure dimensions

            kwargs: keyword arguments for adjacency matrix construction

        """

        self.counts = self.build_matrix(measured, predicted)

        # create figue and add single axis
        if ax is None:
            self.fig = self.create_figure(figsize=figsize)
            self.add_axes()

        else:
            self.fig = plt.gcf()
            self.axes = ax

        # plot classifier adjacency matrix
        self.plot_matrix(self.axes, text=text, **kwargs)

    def save(self, name, dirpath='./', **kwargs):
        """
        Save figure to file.

        Args:

            name (str) - file name without format extension

            dirpath (str) - directory in which to save file

        Keyword arguments:

            fmt (str) - file format, eg 'pdf'

            dpi (int) - resolution

            transparent (bool) - if True, remove background

            rasterized (bool) - if True, rasterize figure data

            kwargs: keyword arguments for plt.savefig

        """
        self._save(self.fig, name, dirpath, **kwargs)

    @staticmethod
    def build_matrix(measured, predicted):
        """
        Compute 2D histogram of measured versus predicted values.

        Args:

            measured (array like) - measured class labels

            predicted (array like) - predicted class labels

        Returns:

            counts (np.ndarray[int]) - pairwise overlap between measured/predicted classes

        """
        counts, _, _ = np.histogram2d(measured, predicted, bins=np.arange(3.5))
        return counts.astype(np.int64)

    def plot_matrix(self, ax, text=None, fontsize=7):
        """
        Plot classification adjacency matrix.

        Args:

            ax (matplotlib.axes.AxesSubplot)

            text (str) - indicates whether 'counts' or 'rates' labels are added

            fontsize (int) - font size for text

        """

        # compute error rates
        rates = self.counts.astype(np.float64)
        rates /= self.counts.sum(axis=1).reshape(-1, 1)

        # plot image
        ax.imshow(rates.T, cmap=plt.cm.Reds, vmin=0, vmax=1)
        ax.invert_yaxis()

        # add labels
        kw = dict(horizontalalignment='center',
                  verticalalignment='center',
                  fontsize=fontsize)

        for i in range(3):
            for j in range(3):

                if i == j:
                    color = 'w'
                    weight = 'bold'
                else:
                    color = 'k'
                    weight ='normal'

                if text=='rates':
                    s = '{:0.0%}'.format(rates[i, j])
                    ax.text(i, j, s=s, color=color, weight=weight, **kw)
                elif text=='counts':
                    s = '{:d}'.format(self.counts[i, j])
                    ax.text(i, j, s=s, color=color, weight=weight, **kw)
                else:
                    continue

        # format axes
        self.format_axis(ax)

    @staticmethod
    def format_axis(ax, labelsize=8):
        """ Format individual axis. """
        ax.set_xlabel('True label', fontsize=labelsize)
        ax.set_ylabel('Assigned label', fontsize=labelsize)
        ax.set_xticks(np.arange(2.5))
        ax.set_xticklabels(['0x', '1x', '2x'], fontsize=labelsize-1)
        ax.set_yticks(np.arange(2.5))
        ax.set_yticklabels(['0x', '1x', '2x'], fontsize=labelsize-1)
        ax.set_aspect(1)



from copy import deepcopy
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize

from .training import Training
from .scoring import Scoring

# import warnings
# warnings.filterwarnings('error')


class BenchmarkProperties:
    """ Properties for Benchmark class. """

    @property
    def xykey(self):
        """ Attribute key for spatial coordinates. """
        return ['x', 'y']

    @property
    def xy(self):
        """ Measurement spatial coordinates. """
        return self.data[self.xykey].values

    @property
    def classifier(self):
        return self.annotator.classifier

    @property
    def fluorescence(self):
        """ Measured fluorescence. """
        return self.data[self.attribute].values

    @property
    def ground_truth(self):
        """ True labels. """
        return self.data.true_dosage.values

    @property
    def labels(self):
        """ Labels assigned by bivariate classifier. """
        return self.data.labels.values

    @property
    def level_only(self):
        """ Labels assigned by univariate fluorescence classifier. """
        return self.data.level_only.values

    @property
    def spatial_only(self):
        """ Labels assigned by univariate spatial classifier. """
        return self.data.spatial_only.values

    @property
    def MAE(self):
        """ Mean absolute error of labels. """
        return self.scores['labels'].MAE

    @property
    def MAE_comm(self):
        """ Mean absolute error of labels when classifier is applied to communities. """
        return self.scores['labels_comm'].MAE

    @property
    def MAE_levels(self):
        """ Mean absolute error of level-only labels. """
        return self.scores['level_only'].MAE

    @property
    def MAE_spatial(self):
        """ Mean absolute error of space-only labels. """
        return self.scores['spatial_only'].MAE


class BenchmarkVisualization:
    """ Visualization methods for Benchmark class. """

    @property
    def norm(self):
        """ Diploid normalization. """
        return Normalize(vmin=0, vmax=2)

    @property
    def fnorm(self):
        """ Fluorescence normalization. """
        return Normalize(*self.annotator.classifier.model.bounds)

    def _scatter(self, ax, c,
                 s=5,
                 cmap=plt.cm.viridis,
                 norm=None,
                 title=None,
                 **kwargs):
        """ Scatter cells in space. """
        if norm is None:
            norm = self.norm
        colors = cmap(norm(c))
        ax.scatter(*self.xy.T, c=colors, s=s, lw=0, **kwargs)
        ax.set_aspect(1)
        ax.axis('off')
        if title is not None:
            ax.set_title(title, fontsize=12)

    def plot_measurements(self, ax=None, norm=None, **kwargs):
        """ Scatter plot of measurement data. """
        if ax is None:
            fig, ax = plt.subplots(figsize=(2, 2))
        c = self.fluorescence

        # define normalization
        if norm is None:
            norm = self.fnorm

        self._scatter(ax, c=c, norm=norm, title=None, **kwargs)

    def show(self, **kwargs):
        """ Plot visual comparison of cell classifiers. """
        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(6, 6))
        self._scatter(axes[0, 0], c=self.ground_truth, title='Ground Truth', **kwargs)
        self._scatter(axes[0, 1], c=self.fluorescence, norm=self.fnorm, title='Fluorescence', **kwargs)
        self._scatter(axes[1, 0], c=self.level_only, title='Level only', **kwargs)
        self._scatter(axes[1, 1], c=self.labels, title='Assigned labels', **kwargs)

    def show_classifiers(self, **kwargs):
        """ Plot visual comparison of cell classifiers. """
        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(6, 6))
        self._scatter(axes[0, 0], c=self.ground_truth, title='Ground Truth', **kwargs)
        self._scatter(axes[0, 1], c=self.labels, title='Assigned labels', **kwargs)
        self._scatter(axes[1, 0], c=self.level_only, title='Level only', **kwargs)
        self._scatter(axes[1, 1], c=self.spatial_only, title='Spatial only', **kwargs)

    def show_measurements(self, **kwargs):
        """ Plot visual comparison of genotypes and measurements. """
        fig, axes = plt.subplots(ncols=2, figsize=(6, 3))
        self._scatter(axes[0], c=self.ground_truth, title='Ground Truth', **kwargs)
        self._scatter(axes[1], c=self.fluorescence, norm=self.fnorm, title='Fluorescence', **kwargs)

    def show_comparison(self, **kwargs):
        """ Plot visual comparison of cell classifiers. """
        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(6, 6))
        self._scatter(axes[0, 0], c=self.level_only, title='Level only', **kwargs)
        self._scatter(axes[1, 0], c=self.labels, title='Assigned labels', **kwargs)
        self.scores['level_only'].plot_matrix(ax=axes[0, 1])
        self.scores['labels'].plot_matrix(ax=axes[1, 1])


class SimulationBenchmark(Training,
                          BenchmarkProperties,
                          BenchmarkVisualization):
    """
    Class for benchmarking a synthetic simulation.

    Attributes:

        data (pd.DataFrame) - synthetic measurement data

        attribute (str) - attribute on which cell measurements are classified

        annotator (Annotation) - object that assigns labels to measurements

    """

    def __init__(self, measurements,
                 annotator=None,
                 graph=None,
                 attribute='clonal_marker',
                 logratio=True,
                 training_kw={},
                 testing_kw={}):
        """
        Args:

            measurements (pd.DataFrame) - synthetic measurement data

            annotator (Annotation) - if None, fit annotator to measurements

            graph (WeightedGraph) - if None, create a new graph

            attribute (str) - attribute on which measurements are classified

            logratio (bool) - if True, weight graph edges by log-ratio of attribute level. otherwise, use the absolute difference

            training_kw (dict) - keyword arguments for annotator training

            testing_kw (dict) - keyword arguments for annotator application

        """

        self.attribute = attribute

        # build graph
        if graph is None:
            graph = self.build_graph(measurements, attribute, logratio)

        # train annotator
        if annotator is None:
            annotator = self.train(graph, attribute=attribute, **training_kw)
        self.annotator = annotator

        # apply graph-based annotation
        measurements['labels'] = annotator.annotate(graph, **testing_kw)

        # apply graph-based annotation based on community
        community_kw = deepcopy(testing_kw)
        community_kw['sampler_type'] = 'community'
        community_kw['sampler_kwargs'] = {'depth': 1.}
        measurements['labels_comm'] = annotator.annotate(graph, **community_kw)

        # apply univariate annotation using only fluorescence levels
        measurements['level_only'] = annotator.classifier[0](measurements)

        # apply bivariate annotation using only spatial context
        bv_kwargs = deepcopy(testing_kw)
        bv_kwargs['bivariate_only'] = True
        measurements['spatial_only'] = annotator.annotate(graph, **bv_kwargs)

        # store measurements
        self.data = measurements

        # score annotation performance
        self.scores = {}
        self.scores['labels'] = self.score(self.labels)
        self.scores['labels_comm'] = self.score(self.data.labels_comm.values)
        self.scores['level_only'] = self.score(self.level_only)
        self.scores['spatial_only'] = self.score(self.spatial_only)

    def score(self, labels):
        """ Assess accuracy of <labels>. """
        return Scoring(self.ground_truth, labels)



from unittest.mock import patch

from os.path import join, exists
import numpy as np

from ..data import Stack
from ..data.images import ImageScalar
from ..bleedthrough.correction import LayerCorrection

from .test_io import TestPaths


class Test03_Layer(TestPaths):
    """
    Tests for Layer instance.
    """

    annotated_channel = 'ch0'
    label_name = 'TEST'

    @classmethod
    def setUpClass(cls):
        """ Initialize test instance of Stack. """
        with patch('builtins.input', side_effect=['yes', 12]):
            stack = Stack(cls.stack_path, bit_depth=cls.bit_depth)
        stack.load_image()
        cls.stack = stack
        cls.layer = cls.stack[0]

    @classmethod
    def tearDownClass(cls):
        """ Restore stack directory to original state. """
        cls.stack.restore_directory()

    def test00_import_segmentation_mask(self):
        """ Import a segmentation mask and generate measurements. """
        self.layer.import_segmentation_mask(self.segmentation_mask_path, 2)
        self.assertTrue(self.layer.is_segmented)

    def test01_import_roi_mask(self):
        """ Define region of interest by importing an ROI mask. """
        self.layer.import_roi_mask(self.roi_mask_path, save=True)
        self.assertTrue('selected' in self.layer.data.columns)

    def test02_build_graph(self):
        """ Build cell adjacency graph. """
        self.layer.build_graph(self.annotated_channel)
        self.assertTrue(self.layer.graph is not None)

    def test03_train_annotator(self):
        """ Train clone annotation model on individual layer. """
        self.layer.train_annotator(self.annotated_channel)
        self.assertTrue(self.layer.annotator is not None)

    def test04_apply_annotator(self):
        """ Annotate layer. """
        self.layer.apply_annotation(label=self.label_name)
        self.assertTrue(self.label_name in self.layer.data.columns)

    def test05_mark_label_boundaries(self):
        """ Mark boundaries between regions with different labels. """
        self.layer.mark_boundaries(self.label_name, max_edges=1)
        self.assertTrue('boundary' in self.layer.data.columns)

    def test06_mark_label_regions(self):
        """ Mark regions in which each label is found. """
        self.layer.apply_concurrency(basis=self.label_name)
        label = list(self.layer.data[self.label_name].unique())[0]
        key = 'concurrent_{:d}'.format(label)
        self.assertTrue(key in self.layer.data.columns)

    def test07_build_attribute_mask(self):
        """ Build attribute mask for layer. """
        mask = self.layer.build_attribute_mask(self.label_name)
        self.assertTrue(type(mask) == np.ma.core.MaskedArray)

    def test08_show_annotation(self):
        """ Show layer annotation. """
        if self.layer.include:
            channel, label = self.annotated_channel, self.label_name
            try:
                fig = self.layer.show_annotation(channel, label)
                success = True
            except:
                success = False
            self.assertTrue(success)

    def test09_select_channel(self):
        """ Select individual fluorescence channel. """
        channel = self.layer.get_channel(1)
        self.assertTrue(type(channel) == ImageScalar)

    def test10_bleedthrough_correction(self):
        """ Correct measurements for fluorescence bleedthrough. """
        correction = LayerCorrection(self.layer, xvar=0, yvar=1)
        correction.save()
        self.assertTrue('ch1c' in self.layer.data.columns)



from os.path import join, isdir, exists
from os import listdir, mkdir
from shutil import rmtree
import gc
from copy import deepcopy
import pandas as pd
import numpy as np
from scipy.ndimage import binary_erosion
import matplotlib.pyplot as plt
from matplotlib.path import Path
from collections import Counter

from ..visualization import *
from ..utilities import IO

# import measurement objects
from ..measurement import Segmentation
from ..measurement import Measurements

# import annotation objects
from ..annotation import WeightedGraph
from ..annotation import Annotation
from ..annotation import ConcurrencyLabeler
from ..annotation import CloneBoundaries
from ..annotation import CelltypeLabeler

# import bleedthrough correction objects
from ..bleedthrough import LayerCorrection

# import image base class
from .images import ImageMultichromatic
from .silhouette_write import WriteSilhouetteLayer

# import default parameters
from .defaults import Defaults
defaults = Defaults()


class LayerVisualization:
    """ Methods for visualizing a layer. """

    @default_figure
    def plot_graph(self, channel,
                   figsize=(15, 15),
                   image_kw={},
                   graph_kw={},
                   ax=None):
        """
        Plot graph on top of relevant image channel.

        Args:

            channel (str) - fluorescence channel to visualize

            figsize (tuple) - figure size

            image_kw (dict) - keyword arguments for scalar image visualization

            graph_kw (dict) - keyword arguments for scalar image visualization

        """

        # add image
        if channel is not None:
            image = self.get_channel(channel)
            image.show(ax=ax, segments=False, **image_kw)

        # add graph
        self.graph.show(ax=ax, **graph_kw)

    def plot_boundary(self, ax,
                        label,
                        label_by='genotype',
                        color='r',
                        alpha=70,
                        **kwargs):
        """ Plot boundary of <label_by> groups with <label> on <ax>. """

        # add labels to ephemeral copy of graph data
        graph = self.graph.copy()
        graph.data[label_by] = self.data[label_by]

        # plot clone boundaries
        bounds = CloneBoundaries(graph, label_by=label_by, alpha=alpha)
        bounds.plot_boundary(label, color=color, ax=ax, **kwargs)

    def plot_boundaries(self, ax,
                        label_by='genotype',
                        cmap=plt.cm.bwr,
                        alpha=70,
                        **kwargs):
        """ Plot boundaries of all <label_by> groups on <ax>. """

        # add labels to ephemeral copy of graph data
        graph = self.graph.copy()
        graph.data[label_by] = self.data[label_by]

        # plot clone boundaries
        bounds = CloneBoundaries(graph, label_by=label_by, alpha=alpha)
        bounds.plot_boundaries(cmap=cmap, ax=ax, **kwargs)

    def _build_mask(self, values,
                   interior_only=False,
                   selection_only=False,
                   null_value=-1):
        """
        Use <values> to construct an image mask.

        Args:

            values (array like) - value/label for each segment

            interior_only (bool) - if True, excludes clone borders

            selection_only (bool) - if True, only include selected region

            null_value (int) - value used to fill unused pixels

        Returns:

            mask (np.ma.Maskedarray) - masked image in which foreground segments are replaced with the specified values

        """

        # build dictionary mapping segments to values
        segment_to_value = dict(zip(self.data.segment_id, values))
        segment_to_value[0] = null_value

        # exclude borders
        if interior_only:
            msg = 'Boundary attribute not found. Annotate and try again.'
            assert 'boundary' in self.data.keys(), msg
            boundary = self.data[self.data.boundary]
            boundary_to_black = {x: null_value for x in boundary.segment_id}
            segment_to_value.update(boundary_to_black)

        # exclude cells not included in selection
        if selection_only:
            excluded = self.data[~self.data.selected]
            excluded_to_black = {x: null_value for x in excluded.id}
            segment_to_value.update(excluded_to_black)

        # construct mask
        segment_to_value = np.vectorize(segment_to_value.get)
        mask = segment_to_value(self.labels)
        mask = np.ma.MaskedArray(mask, mask==null_value)

        return mask

    def build_attribute_mask(self, attribute,
                             interior_only=False,
                             selection_only=False,
                             **kwargs):
        """
        Use <attribute> value for each segment to construct an image mask.

        Args:

            attribute (str) - attribute used to label each segment

            interior_only (bool) - if True, excludes clone borders

            selection_only (bool) - if True, only include selected region

        Returns:

            mask (np.ma.Maskedarray) - masked image in which foreground segments are replaced with the attribute values

        """

        return self._build_mask(self.data[attribute].values,
                                interior_only=interior_only,
                                selection_only=selection_only,
                                **kwargs)

    def build_classifier_mask(self, classifier,
                   interior_only=False,
                   selection_only=False,
                   **kwargs):
        """
        Use segment <classifier> to construct an image mask.

        Args:

            classifier (annotation.Classifier object)

            interior_only (bool) - if True, excludes clone borders

            selection_only (bool) - if True, only include selected region

        Returns:

            mask (np.ma.Maskedarray) - masked image in which foreground segments are replaced with the assigned labels

        """
        return self._build_mask(classifier(self.data),
                                interior_only=interior_only,
                                selection_only=selection_only,
                                **kwargs)


class LayerIO(WriteSilhouetteLayer):
    """
    Methods for saving and loading Layer objects and their subcomponents.
    """

    def make_subdir(self, dirname):
        """ Make subdirectory. """
        dirpath = join(self.path, dirname)
        if not exists(dirpath):
            mkdir(dirpath)
        self.add_subdir(dirname, dirpath)

    def add_subdir(self, dirname, dirpath):
        """ Add subdirectory. """
        self.subdirs[dirname] = dirpath

    def find_subdirs(self):
        """ Find all subdirectories. """
        self.subdirs = {}
        for dirname in listdir(self.path):
            dirpath = join(self.path, dirname)
            if isdir(dirpath):
                self.add_subdir(dirname, dirpath)

    def save_metadata(self):
        """ Save metadata. """
        io = IO()
        io.write_json(join(self.path, 'metadata.json'), self.metadata)

    def save_segmentation(self, image, **kwargs):
        """
        Save segment labels, and optionally save a segmentation image.

        Args:

            image (bool) - if True, save segmentation image

            kwargs: keyword arguments for image rendering

        """
        dirpath = self.subdirs['segmentation']

        # save segment labels
        np.save(join(dirpath, 'labels.npy'), self.labels)

        # save segmentation image
        if image:
            bg = self.get_channel(self.metadata['bg'], copy=False)
            fig = bg.show(segments=True)
            fig.axes[0].axis('off')
            fig.savefig(join(dirpath, 'segmentation.png'), **kwargs)
            fig.clf()
            plt.close(fig)
            gc.collect()

    def save_measurements(self):
        """ Save raw measurements. """

        # get segmentation directory
        path = join(self.subdirs['measurements'], 'measurements.hdf')

        # save raw measurements
        self.measurements.to_hdf(path, 'measurements', mode='w')

    def save_processed_data(self):
        """ Save processed measurement data. """

        path = join(self.subdirs['measurements'], 'processed.hdf')
        self.data.to_hdf(path, 'data', mode='w')

    def save_annotator(self, image=False, **kwargs):
        """
        Save annotator instance.

        Args:

            image (bool) - if True, save annotation images

            kwargs: keyword arguments for image rendering

        """
        path = self.subdirs['annotation']
        self.annotator.save(path, image=image, **kwargs)

    def save(self,
             segmentation=True,
             measurements=True,
             processed_data=True,
             annotator=False,
             segmentation_image=False,
             annotation_image=False):
        """
        Save segmentation parameters and results.

        Args:

            segmentation (bool) - if True, save segmentation

            measurements (bool) - if True, save measurement data

            processed_data (bool) - if True, save processed measurement data

            annotator (bool) - if True, save annotator

            segmentation_image (bool) - if True, save segmentation image

            annotation_image (bool) - if True, save annotation image

        """

        # set image keyword arguments
        image_kw = dict(format='png',
                     dpi=100,
                     bbox_inches='tight',
                     pad_inches=0,
                     transparent=True,
                     rasterized=True)

        # save segmentation
        if segmentation:
            self.make_subdir('segmentation')
            self.save_segmentation(image=segmentation_image, **image_kw)

        # save measurements
        if measurements:
            self.make_subdir('measurements')
            self.save_measurements()

        # save processed data
        if processed_data and self.data is not None:
            self.data = self.process_measurements(self.measurements)
            self.save_processed_data()

        # save annotation
        if annotator and self.annotator is not None:
            self.make_subdir('annotation')
            self.save_annotator(image=annotation_image, **image_kw)

        # save metadata
        self.save_metadata()

    def load_metadata(self):
        """ Load metadata. """
        path = join(self.path, 'metadata.json')
        if exists(path):
            io = IO()
            self.metadata = io.read_json(path)

    def load_labels(self):
        """ Load segment labels if they are available. """
        labels = None
        if 'segmentation' in self.subdirs.keys():
            segmentation_path = self.subdirs['segmentation']
            labels_path = join(segmentation_path, 'labels.npy')
            if exists(labels_path):
                labels = np.load(labels_path)
        self.labels = labels

    def load_measurements(self):
        """ Load raw measurements. """
        path = join(self.subdirs['measurements'], 'measurements.hdf')
        self.measurements = pd.read_hdf(path, 'measurements')

    def load_processed_data(self):
        """ Load processed data from file. """
        path = join(self.subdirs['measurements'], 'processed.hdf')
        self.data = pd.read_hdf(path, 'data')

    def load_annotator(self):
        """ Load annotator instance. """
        self.annotator = Annotation.load(self.subdirs['annotation'])

    def load_inclusion(self):
        """ Load inclusion flag. """
        io = IO()
        selection_md = io.read_json(join(self.subdirs['selection'], 'md.json'))
        if selection_md is not None:
            self.include = bool(selection_md['include'])

    def load_correction(self):
        """
        Load linear background correction.

        Returns:

           correction (LayerCorrection)

        """
        assert self.has_image, 'Image unavailable. Load image and try again.'
        return LayerCorrection.load(self)

    def load(self, use_cache=True, graph=True):
        """
        Load layer.

        Args:

            use_cache (bool) - if True, use cached measurement data, otherwise re-process the measurement data

            graph (bool) - if True, load weighted graph

        """

        # load metadata and extract background channel
        self.load_metadata()

        # load inclusion data
        if 'selection' in self.subdirs.keys():
            self.load_inclusion()

        # if layer is not included, skip it
        if not self.include:
            return None

        # check whether annotation exists
        if 'annotation' in self.subdirs.keys() and not use_cache:

            if self.annotator is not None:
                raise UserWarning('Layer was instantiated with a stack-level annotation instance, but a second annotation instance was found within the layer directory. Resolve this conflict before continuing.')

            # load annotator
            self.load_annotator()

        # check whether segmentation exists and load raw measurement data
        if 'measurements' in self.subdirs.keys():
            self.load_measurements()

        # if processing measurements, ensure that graph is built
        if not use_cache:
            graph = True

        # build graph
        if graph and 'graph_weighted_by' in self.metadata['params'].keys():
            graph_weighted_by = self.metadata['params']['graph_weighted_by']
            graph_kw = self.metadata['params']['graph_kw']
            self.build_graph(graph_weighted_by, **graph_kw)
        else:
            self.graph = None

        # check whether cached measurements are available
        if 'measurements' in self.subdirs.keys():
            path = join(self.subdirs['measurements'], 'processed.hdf')

            # load processed data
            if use_cache and exists(path):
                self.load_processed_data()

            # otherwise, process raw measurement data
            else:
                self.data = self.process_measurements(self.measurements)


class LayerProperties:
    """
    Properties for Layer class:

        color_depth (int) - number of fluorescence channels

        num_cells (int) - number of cells detected by segmentation

        bg_key (str) - key for channel used to generate segmentation

        has_image (bool) - if True, image is loaded into memory

        is_segmented (bool) - if True, layer has been segmented

        has_trained_annotator (bool) - if True, layer has a trained annotator

    """

    @property
    def color_depth(self):
        """ Number of color channels. """
        return self.im.shape[-1]

    @property
    def num_cells(self):
        """ Number of cells detected by segmentation. """
        return len(self.data) if self.data is not None else None

    @property
    def bg_key(self):
        """ DataFrame key for background channel. """
        return self._to_key(self.metadata['bg'])

    @property
    def has_image(self):
        """ True if image is available. """
        return self.im is not None

    @property
    def is_segmented(self):
        """ True if measurement data are available. """
        return self.measurements is not None

    @property
    def has_trained_annotator(self):
        """ Returns True if trained annotator is available. """
        return self.annotator is not None


class LayerMeasurement:
    """

    Measurement related methods for Layer class.

    """

    def segment(self, channel,
                preprocessing_kws={},
                seed_kws={},
                seg_kws={},
                min_area=250):
        """
        Identify nuclear contours by running watershed segmentation on specified background channel.

        Args:

            channel (int) - channel index on which to segment image

            preprocessing_kws (dict) - keyword arguments for image preprocessing

            seed_kws (dict) - keyword arguments for seed detection

            seg_kws (dict) - keyword arguments for segmentation

            min_area (int) - threshold for minimum segment size, px

        Returns:

            background (ImageScalar) - background image (after processing)

        """

        # append default parameter values
        preprocessing_kws = defaults('preprocessing', preprocessing_kws)
        seed_kws = defaults('seeds', seed_kws)
        seg_kws = defaults('segmentation', seg_kws)

        # store parameters in metadata
        self.metadata['bg'] = channel
        segmentation_kw = dict(preprocessing_kws=preprocessing_kws,
                               seed_kws=seed_kws,
                               seg_kws=seg_kws,
                               min_area=min_area,
                               imported=False)
        self.metadata['params']['segmentation_kw'] = segmentation_kw

        # extract and preprocess background
        background = self.get_channel(channel)
        background.preprocess(**preprocessing_kws)

        # run segmentation
        seg = Segmentation(background, seed_kws=seed_kws, seg_kws=seg_kws)

        # exclude small segments
        seg.exclude_small_segments(min_area=min_area)

        # update segment labels
        self.labels = seg.labels
        background.labels = seg.labels

        # update cell measurements
        self.measure()

        return background

    def measure(self):
        """
        Measure properties of cell segments. Raw measurements are stored under in the 'measurements' attribute, while processed measurements are stored in the 'data' attribute.
        """

        # measure segment properties
        measurements = Measurements(self.im, self.labels)
        measurements = measurements.build_dataframe()

        # assign layer id, apply normalization, and save measurements
        measurements['layer'] = self._id
        self.apply_normalization(measurements)
        self.measurements = measurements

        # process raw measurement data
        self.data = self.process_measurements(measurements)

    def apply_normalization(self, data):
        """
        Normalize fluorescence intensity measurements by measured background channel intensity.

        Args:

            data (pd.DataFrame) - processed cell measurement data

        """

        # get background channel from metadata
        bg = self.metadata['bg']

        # apply normalization to each foreground channel
        for fg in range(self.color_depth):
            if fg == bg:
                continue
            fg_key = self._to_key(fg)
            data['{:s}_normalized'.format(fg_key)] = data[fg_key]/data[self.bg_key]

    def import_segmentation_mask(self, path, channel,
                                save=True,
                                save_image=True):
            """
            Import external segmentation mask and use it to generate measurements.

            Provided mask must contain a 2-D array of positive integers in which a values of zero denotes the image background.

            Args:

                path (str) - path to segmentation mask

                channel (int) - fluorescence channel used for segmentation

                save (bool) - if True, copy segmentation to stack directory

                save_image (bool) - if True, save segmentation image

            """

            assert exists(path), 'File does not exist.'

            io = IO()
            mask = io.read_npy(path)

            int_types = (int, np.int32, np.int64)
            assert mask.dtype in int_types, 'Mask does not contain integers.'
            assert mask.shape == self.shape, 'Mask dimensions are incorrect.'
            assert mask.min() >= 0, 'Mask contains values less than zero.'

            # set segmentation mask and generate measurements
            self.labels = mask
            self.metadata['bg'] = channel
            self.measure()

            # optionally copy mask to stack directory
            if save:
                self.metadata['params']['segmentation_kw']=dict(imported=True)
                self.save_metadata()

                self.make_subdir('segmentation')
                self.save_segmentation(save_image)

                self.make_subdir('measurements')
                self.save_measurements()


class LayerROI:
    """

    ROI related methods for Layer class.

    """

    @staticmethod
    def _apply_roi_vertices(data, xykey, roi_vertices):
        """
        Label cells within a specified region of interest.

        Args:

            data (pd.DataFrame) - cell measurement data

            roi_vertices (np.ndarray[int], N x 2) - vertices bounding ROI

        """

        # add selected attribute to cell measurement data
        data['selected'] = False

        # construct matplotlib path object
        path = Path(roi_vertices, closed=False)

        # mark cells as within or outside the selection boundary
        xy_positions = data[xykey].values
        data['selected'] = path.contains_points(xy_positions)

    @staticmethod
    def sort_clockwise(xycoords):
        """ Returns clockwise-sorted xy coordinates. """
        return xycoords[:, np.argsort(np.arctan2(*(xycoords.T - xycoords.mean(axis=1)).T))]

    @classmethod
    def mask_to_vertices(cls, mask):
        """
        Convert boolean mask to a list of vertices defining the border around the largest contiguous region.

        Args:

            mask (np.ndarray[bool]) - ROI mask, where True denotes the region. Note that the mask may only contain one contiguous component.

        Returns:

            vertices (np.ndarray[int]) - N x 2 array of vertices

        """

        borders = (mask != binary_erosion(mask, structure=np.ones((3, 3))))
        vertices = cls.sort_clockwise(np.asarray(borders.nonzero()))
        return vertices.T

    def import_roi_mask(self, path, save=True):
        """
        Import external ROI mask and use it to label measurement data.

        Provided mask must contain a 2-D boolean array with the same dimensions as the raw image. True values denote the ROI. The mask may only contain a single contiguous ROI.

        Args:

            path (str) - path to ROI mask

            save (bool) - if True, copy ROI mask to stack directory

        """

        assert exists(path), 'File does not exist.'

        # read mask and make sure it's valid
        io = IO()
        mask = io.read_npy(path)
        assert mask.min()>=0 and mask.max()<=1, 'Mask is not boolean.'
        assert mask.shape == self.shape, 'Mask dimensions are incorrect.'
        mask = mask.astype(bool)

        # convert mask to vertices and apply to measurement data
        vertices = self.mask_to_vertices(mask)
        self._apply_roi_vertices(self.data, self.xykey, vertices)

        # save ROI mask to stack directory
        if save:
            self.make_subdir('selection')
            selection_path = self.subdirs['selection']
            io = IO()
            io.write_npy(join(selection_path, 'selection.npy'), vertices)
            md = dict(include=True)
            io.write_json(join(selection_path, 'md.json'), md)

            # update measurements
            self.save_processed_data()

    def define_roi(self, data):
        """
        Adds a "selected" attribute to measurements dataframe. The attribute is True for cells that fall within the ROI.

        Args:

            data (pd.DataFrame) - processed measurement data

        """

        if self.include:

            # load ROI vertices
            io = IO()
            path = join(self.subdirs['selection'],'selection.npy')
            roi_vertices = io.read_npy(path)

            # apply mask
            self._apply_roi_vertices(data, self.xykey, roi_vertices)

        else:
            data['selected'] = False


class LayerCorrection:
    """

    Bleedthrough correction related methods for Layer class.

    """

    def apply_correction(self, data):
        """
        Adds bleedthrough-corrected fluorescence levels to the measurements dataframe.

        Args:

            data (pd.DataFrame) - processed cell measurement data

        """

        # load correction coefficients and X/Y variables
        io = IO()
        cdata = io.read_json(join(self.subdirs['correction'], 'data.json'))

        # get independent/dependent variables
        xvar = cdata['params']['xvar']
        yvar = cdata['params']['yvar']
        bgvar = self.metadata['bg']
        if type(xvar) == int:
            xvar = 'ch{:d}'.format(xvar)
        if type(yvar) == int:
            yvar = 'ch{:d}'.format(yvar)
        if type(bgvar) == int:
            bgvar = 'ch{:d}'.format(bgvar)

        # get linear model coefficients
        b, m = cdata['coefficients']

        # apply correction
        trend = b + m * data[xvar].values
        data[yvar+'_predicted'] = trend
        data[yvar+'c'] = data[yvar] - trend
        data[yvar+'c_normalized'] = data[yvar+'c'] / data[bgvar]


class LayerAnnotation:
    """

    Annotation related methods for Layer class.

    """

    def annotate(self):
        """
        Annotate measurement data in place, also labeling boundaries between labeled regions and marking regions in which each label occurs.
        """

        # make sure graph is available
        msg = 'Graph not found. Call the .build_graph() method then try again.'
        assert self.graph is not None, msg

        # make sure annotator is available
        msg = 'Trained annotator not found. Call the .train_annotator() method then try again.'
        assert self.has_trained_annotator, msg

        # apply trained annotator to label distinct celltypes
        self._apply_annotation(self.data)

        # mark boundaries between labeled regions
        self._mark_boundaries(self.data, basis='genotype', max_edges=1)

        # mark regions in which each label is found
        self._apply_concurrency(self.data, basis='genotype')

    def train_annotator(self, attribute,
                        save=False,
                        logratio=True,
                        num_labels=3,
                        **kwargs):
        """
        Train an Annotation model on the measurements in this layer.

        Args:

            attribute (str) - measured attribute used to determine labels

            save (bool) - if True, save model selection routine

            logratio (bool) - if True, weight edges by relative attribute value

            num_labels (int) - number of allowable unique labels

            kwargs: keyword arguments for Annotation, including:

                sampler_type (str) - either 'radial', 'neighbors', 'community'

                sampler_kwargs (dict) - keyword arguments for sampler

                min_num_components (int) - minimum number of mixture components

                max_num_components (int) - maximum number of mixture components

                addtl_kwargs: keyword arguments for Classifier

        Returns:

            selector (ModelSelection object)

        """

        # instantiate annotator
        self.annotator = Annotation(attribute, num_labels=num_labels, **kwargs)

        # build graph and use it to train annotator
        self.build_graph(attribute, logratio=logratio)
        selector = self.annotator.train(self.graph)

        # save trained annotator
        if save:
            self.save_metadata()
            self.make_subdir('annotation')
            selector.save(self.subdirs['annotation'])

        return selector

    def _apply_annotation(self, data,
                          label='genotype',
                          **kwargs):
        """
        Assign labels to cell measurements.

        Args:

            data (pd.DataFrame) - processed cell measurement data

            label (str) - attribute name for predicted genotype

            kwargs: keyword arguments for Annotator.annotate()

        """
        data[label] = self.annotator(self.graph, **kwargs)

    def apply_annotation(self, label='genotype', **kwargs):
        """
        Assign labels to cell measurements in place.

        Args:

            label (str) - attribute name for predicted genotype

            kwargs: keyword arguments for Annotator.annotate()

        """
        self._apply_annotation(self.data, label=label, **kwargs)

    @staticmethod
    def _apply_concurrency(data, basis='genotype',
                          min_pop=5,
                          max_distance=10,
                          **kwargs):
        """
        Add boolean 'concurrent_<basis>' field to measurement data for each unique value of <basis> attribute.

        Args:

            data (pd.DataFrame) - processed cell measurement data

            basis (str) - attribute on which concurrency is established

            min_pop (int) - minimum population size for inclusion of cell type

            max_distance (float) - maximum distance threshold for inclusion

            kwargs: keyword arguments for ConcurrencyLabeler

        """

        assert basis in data.columns, 'Attribute {:s} not found.'.format(basis)

        labeler = ConcurrencyLabeler(attribute=basis,
                                    min_pop=min_pop,
                                    max_distance=max_distance,
                                    **kwargs)
        labeler(data)

    def apply_concurrency(self, basis='genotype',
                          min_pop=5,
                          max_distance=10,
                          **kwargs):
        """
        Add boolean 'concurrent_<basis>' field to measurement data for each unique value of <basis> attribute.

        Args:

            basis (str) - attribute on which concurrency is established

            min_pop (int) - minimum population size for inclusion of cell type

            max_distance (float) - maximum distance threshold for inclusion

            kwargs: keyword arguments for ConcurrencyLabeler

        """

        self._apply_concurrency(self.data,
                                basis=basis,
                                min_pop=min_pop,
                                max_distance=max_distance,
                                **kwargs)

    def _mark_boundaries(self, data, basis='genotype', max_edges=0):
        """
        Mark boundaries between cells with disparate labels by assigning a boundary label to all cells that share an edge with another cell with a different label.

        Args:

            data (pd.DataFrame) - processed cell measurement data

            basis (str) - attribute used to define label

            max_edges (int) - maximum number of edges for interior cells

        """

        # make sure graph is available
        msg = 'Graph not found, call .build_graph() method then try again.'
        assert self.graph is not None, msg

        # make sure basis attribute is available
        msg = 'Attribute {:s} not found in measurement data.'.format(basis)
        assert basis in data.columns, msg

        # assign genotype to edges
        assign_genotype = np.vectorize(dict(data[basis]).get)
        edge_genotypes = assign_genotype(self.graph.edges)

        # find edges traversing clones
        boundaries = (edge_genotypes[:, 0] != edge_genotypes[:, 1])

        # get number of clone-traversing edges per node
        boundary_edges = self.graph.edges[boundaries]
        edge_counts = Counter(boundary_edges.ravel())

        # assign boundary label to nodes with too many clone-traversing edges
        boundary_nodes = [n for n, c in edge_counts.items() if c>max_edges]
        data['boundary'] = False
        data.loc[boundary_nodes, 'boundary'] = True

    def mark_boundaries(self, basis='genotype', max_edges=0):
        """
        Mark boundaries between cells with disparate labels by assigning a boundary label to all cells that share an edge with another cell with a different label.

        Args:

            basis (str) - attribute used to define label

            max_edges (int) - maximum number of edges for interior cells

        """
        self._mark_boundaries(self.data, basis=basis, max_edges=max_edges)

    def show_annotation(self, channel, label,
                        interior_only=False,
                        selection_only=False,
                        cmap=None,
                        figsize=(8, 4),
                        **kwargs):
        """

        Visualize annotation by overlaying <label> attribute on the image of the specified fluoreascence <channel>.

        Args:

            channel (str) - fluorescence channel to visualize

            label (str) - attribute containing cell type labels

            interior_only (bool) - if True, exclude border regions

            selection_only (bool) - if True, only add contours within ROI

            cmap (matplotlib.ListedColorMap) - color scheme for celltype labels

            figsize (tuple) - figure dimensions

            kwargs: keyword arguments for plt.scatter

        Returns:

            fig (matplotlib.Figure)

        """

        assert label in self.data.keys(), 'No {:s} attribute found. Please check to make sure that annotation is complete.'.format(label)

        # create figure and plot images
        fig, (ax0, ax1) = plt.subplots(ncols=2, figsize=figsize)
        _ = self.get_channel(channel).show(segments=False, ax=ax0)
        _ = self.get_channel(channel).show(segments=False, ax=ax1)

        # build and overlay attribute mask
        mask = self.build_attribute_mask(label,
                                         interior_only=interior_only,
                                         selection_only=selection_only)
        ax1.imshow(mask, cmap=cmap)

        # rectify dimensions
        ax1.set_xlim(*ax0.get_xlim())
        ax1.set_ylim(*ax0.get_ylim())
        plt.tight_layout()

        return fig


class Layer(LayerIO,
            ImageMultichromatic,
            LayerVisualization,
            LayerProperties,
            LayerMeasurement,
            LayerROI,
            LayerCorrection,
            LayerAnnotation):
    """
    Object represents a single imaged layer.

    Attributes:

        measurements (pd.DataFrame) - raw cell measurement data

        data (pd.DataFrame) - processed cell measurement data

        path (str) - path to layer directory

        _id (int) - layer ID, must be an integer value

        subdirs (dict) - {name: path} pairs for all subdirectories

        metadata (dict) - layer metadata

        labels (np.ndarray[int]) - segment ID mask

        annotator (Annotation) - object that assigns labels to measurements

        graph (Graph) - graph connecting cell centroids

        include (bool) - if True, layer was manually marked for inclusion

    Inherited attributes:

        im (np.ndarray[float]) - 3D array of pixel values

        shape (array like) - image dimensions

        mask (np.ndarray[bool]) - image mask

        labels (np.ndarray[int]) - segment ID mask

    Properties:

        color_depth (int) - number of fluorescence channels

        num_cells (int) - number of cells detected by segmentation

        bg_key (str) - key for channel used to generate segmentation

        is_segmented (bool) - if True, layer has been segmented

        has_trained_annotator (bool) - if True, layer has a trained annotator

    """

    def __init__(self, path, im=None, annotator=None):
        """
        Instantiate layer.

        Args:

            path (str) - path to layer directory

            im (np.ndarray[float]) - 3D array of pixel values

            annotator (Annotation) - object that assigns labels to measurements

        """

        # set layer ID
        layer_id = int(path.rsplit('/', maxsplit=1)[-1])
        self._id = layer_id
        self.xykey = ['centroid_x', 'centroid_y']

        # set path and subdirectories
        self.path = path

        # make layers directory
        if not exists(self.path):
            self.initialize()
        self.find_subdirs()

        # load inclusion; defaults to True
        if 'selection' in self.subdirs.keys():
            if len(listdir(self.subdirs['selection'])) == 0:
                self.include = True
            else:
                self.load_inclusion()
        else:
            self.include = True

        # initialize measurement data
        self.measurements = None
        self.data = None

        # set annotator
        self.annotator = annotator

        # load labels and instantiate image
        self.load_labels()
        super().__init__(im, labels=self.labels)

    def initialize(self):
        """

        Initialize layer directory by:

            - Creating a layer directory
            - Removing existing segmentation directory
            - Saving metadata to file

        """

        # make layers directory
        if not exists(self.path):
            mkdir(self.path)
        self.subdirs = {}

        # remove existing segmentation/annotation/measurement directories
        for key in ('segmentation', 'measurements', 'annotation'):
            if key in self.subdirs.keys():
                rmtree(self.subdirs[key])

        # make metadata file
        segmentation_kw = dict(preprocessing_kws={}, seed_kws={}, seg_kws={})
        params = dict(segmentation_kw=segmentation_kw, graph_kw={})
        metadata = dict(bg=None, params=params)

        # save metadata
        IO().write_json(join(self.path, 'metadata.json'), metadata)

    def process_measurements(self, measurements):
        """
        Augment measurements by:
            1. incorporating manual selection boundary
            2. correcting for fluorescence bleedthrough
            3. assigning measurement labels
            4. marking clone boundaries
            5. assigning label concurrency information

        Operations 3-5 require construction of a WeightedGraph object.

        Args:

            measurements (pd.DataFrame) - raw measurement data

        Returns:

            data (pd.DataFrame) - processed measurement data

        """

        # copy raw measurements
        data = deepcopy(measurements)

        # load and apply selection
        if 'selection' in self.subdirs.keys():
            self.define_roi(data)

        # load and apply correction
        if 'correction' in self.subdirs.keys():
            self.apply_correction(data)

        # annotate measurements
        if self.has_trained_annotator and self.graph is not None:

            # apply trained annotator to label distinct celltypes
            self._apply_annotation(data, label='genotype')

            # mark boundaries between labeled regions
            self._mark_boundaries(data, basis='genotype', max_edges=1)

            # mark regions in which each label is found
            self._apply_concurrency(data, basis='genotype')

        return data

    def build_graph(self, weighted_by, **graph_kw):
        """
        Compile weighted graph connecting adjacent cells.

        Args:

            weighted_by (str) - attribute used to weight edges

            graph_kw: keyword arguments, including:

                xykey (list) - attribute keys for node x/y positions

                logratio (bool) - if True, weight edges by log ratio

                distance (bool) - if True, weights edges by distance

        """

        # store metadata for graph reconstruction
        self.metadata['params']['graph_weighted_by'] = weighted_by
        self.metadata['params']['graph_kw'] = graph_kw

        # build graph
        self.graph = WeightedGraph(self.measurements, weighted_by, **graph_kw)



from time import time
from os.path import curdir
from clones.validation.arguments import RunArguments
from clones.validation.batch import BatchBenchmark

# ======================== DEFINE FUNCTIONALITY ===============================

def foo(job):
    """ Function applied to <job>. """
    pass
    # save benchmark
    #job.save(path)


# ======================== PARSE SCRIPT ARGUMENTS =============================

args = RunArguments(description='Batch benchmark arguments.')
job_path = args['path']

# ============================= RUN SCRIPT ====================================

start_time = time()

# run each simulation in job file
with open(job_path, 'r') as job_file:

    # run each simulation
    for path in job_file.readlines():

        # load benchmark job
        job = BatchBenchmark.load(path.strip())
        job.batch.root = curdir

        # DO STUFF (defined by <foo>)
        foo(job)

# print runtime to standard out
runtime = time() - start_time
print('BATCH COMPLETED IN {:0.2f}.\n'.format(runtime))



from time import time
from os.path import curdir
from clones.validation.arguments import RunArguments
from clones.validation.batch import BatchBenchmark


# ======================== PARSE SCRIPT ARGUMENTS =============================

args = RunArguments(description='Batch benchmark arguments.')
path = args['path']
data = args['save_data']

# ============================= RUN SCRIPT ====================================

start_time = time()

# load benchmark
benchmark = BatchBenchmark.load(path)
benchmark.batch.root = curdir

# evaluate benchmark
benchmark.run()

# save benchmark
benchmark.save(path, data=data)

# print runtime to standard out
runtime = time() - start_time
print('BENCHMARK COMPLETED IN {:0.2f}.\n'.format(runtime))



import numpy as np
from scipy.spatial.distance import cdist

from ..labelers import AttributeLabeler


class ConcurrencyLabeler(AttributeLabeler):
    """
    Object for labeling cell measurements as concurrent with other cell types.  Concurrency with each unique target cell type is determined by applying a maximum threshold on the x-distance between each cell and its nearest neighbor of the target type.

    Attributes:

        attribute (str) - attribute used to denote cell type

        label_values (array like) - by default use all available

        min_pop (int) - minimum population size for inclusion of cell type

        max_distance (float) - maximum distance threshold for inclusion

    """

    def __init__(self,
                 attribute='genotype',
                 label_values=None,
                 min_pop=5, max_distance=10):
        """
        Instantiate object for labeling measurements as concurrent with other cell types.

        Args:

            attribute (str) - attribute used to denote labels

            label_values (array like) - by default use all available

            min_pop (int) - minimum population size for inclusion of cell type

            max_distance (float) - maximum distance threshold for inclusion

        """

        self.attribute = attribute
        self.label_values = label_values
        self.min_pop = min_pop
        self.max_distance = max_distance

    def evaluate_distance(self, data, target_type):
        """
        Evaluate x-distance of all cells to nearest neighbor of target cell type.

        Args:

            data (pd.DataFrame) - processed cell measurement data

            target_type (str) - target cell type

        Returns:

            distances (np.ndarray[float]) - distances to target cell type

        """

        # get target cells
        targets = data[data[self.attribute]==target_type]

        # evaluate distances if there are enough target cells
        if len(targets) > self.min_pop:
            rs = lambda x: x.centroid_x.values.reshape(-1, 1)
            distances = cdist(rs(data), rs(targets)).min(axis=1)

        # otherwise set arbitrarily high distance
        else:
            distances = 10000*np.ones(len(data), dtype=np.float64)

        return distances

    def assign_labels(self, data):
        """
        Add boolean 'concurrent_<label>' field to measurement data for each specified label value.

        Args:

            data (pd.DataFrame) - processed cell measurement data

        """

        # find unique target cell types if no label values specified
        if self.label_values is None:
            target_types = data[self.attribute].unique()
        else:
            target_types = self.label_values

        # assign concurrency label
        for target_type in target_types:
            distances = self.evaluate_distance(data, target_type)
            key = 'concurrent_{:s}'.format(str(target_type))
            data[key] = (distances <= self.max_distance)



from clones.validation.arguments import SweepArguments
from clones.validation.sweep import SweepBenchmark

from os.path import join
from glob import glob
from shutil import move


# ======================== PARSE SCRIPT ARGUMENTS =============================

args = SweepArguments(description='SweepBenchmark arguments.')

benchmarking = SweepBenchmark.load(args['path'])


# ============================= EDIT JOB FILES ================================

jobs_path = join(benchmarking.path, 'jobs')
for filepath in glob(join(jobs_path, '[0-9]*.txt')):

    # read existing file contents
    with open(filepath, 'r') as file:
        infile = file.readlines()

    # remove [0-9]*.pkl
    new_lines = [line.strip().rsplit('/', maxsplit=1)[0] for line in infile]

    # write new file contents
    with open(filepath, 'w') as file:
        file.writelines(new_lines)


# ========================= RENAME BATCH FILES ================================

batches_path = join(benchmarking.path, 'batches')
for filepath in glob(join(batches_path, '[0-9]*')):
    src = join(filepath, '{}.pkl'.format(filepath.rsplit('/', maxsplit=1)[-1]))
    dst = join(filepath, 'batch.pkl')
    move(src, dst)








from os.path import join
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import seaborn as sns

from ..visualization import *

from .resampling import resample_uniformly
from .background import BackgroundExtraction


class CorrectionVisualization:
    """
    Methods for visualizing correction procedure.
    """

    def show_fit(self, mode='box', bin_size=0.05, figsize=(3, 2)):
        """ Plot fit to background pixels using sns.boxplot . """

        # intantiate figure
        fig, ax = plt.subplots(figsize=figsize)

        # compile dataframe
        bg_xy = np.vstack((self.x, self.y)).T
        data = pd.DataFrame.from_records(bg_xy, columns=['x', 'y'])

        # add data to plot
        data['bin'] = (data.x // bin_size)
        if mode == 'strip':
            sns.stripplot(x='bin', y='y', data=data, ax=ax, size=1, color='k')
        elif mode == 'box':
            sns.boxplot(x='bin', y='y', data=data, color='grey', ax=ax, width=.6, fliersize=2)
        elif mode == 'violin':
            sns.violinplot(x='bin', y='y', data=data, color='grey', ax=ax, width=.6)
        elif mode == 'scatter':
            sns.scatterplot(x='bin', y='y', data=data)
        elif mode == 'raw':
            ax.plot(data.x, data.y, '.k', markersize=1)

        # set limits
        xlim = (-0.5, data.bin.max()+0.5)
        ax.set_xlim(*xlim)

        # format axes
        ax.set_xlabel('Red intensity')
        ax.set_ylabel('Green intensity')
        xticks = ax.get_xticks() * bin_size
        ax.set_xticklabels(['{:0.2f}'.format(s) for s in xticks])
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)

        # add model prediction to plot
        ax.plot(ax.get_xticks(), self.predict(xticks), '-r', linewidth=1)

        # store figure instance
        self.figs['fit'] = fig

    def show_correction(self, figsize=(6, 2), selected_only=False):
        """
        Show cell measurements before and after correction.

        Args:

            figsize (tuple) - figure size

            selected_only (bool) - if True, exclude cells outside selection bounds

        """

        # instantiate figure
        fig = plt.figure(figsize=figsize)
        gs = GridSpec(nrows=1, ncols=2, wspace=.3)
        ax0 = plt.subplot(gs[0])
        ax1 = plt.subplot(gs[1])

        # add data to plots
        if selected_only:
            mask = self.data.selected.values
        else:
            mask = np.ones(self.xt.size, dtype=bool)

        ax0.scatter(self.xt[mask], self.yt[mask], c='k', s=1, linewidth=0)
        ax1.scatter(self.xt[mask], self.ytc[mask], c='k', s=1, linewidth=0)

        # add model prediction to plot (dashed line for extrapolation)
        ax0.plot(self.xtdomain, self.predict(self.xtdomain), '--r', lw=1)
        ax0.plot(self.domain, self.predict(self.domain), '-r', lw=1.5)

        # label axes
        ax0.set_xlabel('Measured RFP level')
        ax0.set_ylabel('Measured GFP level')
        ax1.set_ylabel('Corrected GFP level')
        ax1.set_xlabel('Measured RFP level')

        # format axes
        xlim = (-0.02, self.xtdomain.max()+0.02)
        ylim = (-0.05, self.yt.max())
        for ax in (ax0, ax1):
            ax.set_xlim(*xlim)
            ax.set_ylim(*ylim)
            ax.spines['top'].set_visible(False)
            ax.spines['right'].set_visible(False)

        # store figure instance
        self.figs['correction'] = fig

    @staticmethod
    def _pixel_distribution(x0, xf, label='', bins=None, **kwargs):
        """
        Plot the background pixel intensity distributions before and after resampling.
        """

        # create figure
        fig = plt.figure(figsize=(3.5, 1.))
        gs = GridSpec(nrows=1, ncols=2, wspace=0.3)
        ax0 = fig.add_subplot(gs[0])
        ax1 = fig.add_subplot(gs[1])
        axes = (ax0, ax1)

        if bins is None:
            bins = np.arange(0, .4, .05)

        # plot pixel intensity distributions
        _ = axes[0].hist(x0, bins=bins, color='k')
        _ = axes[1].hist(xf, bins=bins, color='k')

        # format axes
        for ax in axes:
            ax.set_xlim(0, bins.max())
            ax.spines['top'].set_visible(False)
            ax.spines['right'].set_visible(False)
            #ax.set_yticks([])
            ax.set_ylim(bottom=1, top=2e5)
            ax.set_yscale('symlog', linthreshy=.9, nonposy="clip")
            ax.set_yticks([1, 10, 100, 1000, 10000, 100000])
            ax.set_ylabel('No. pixels')
        ax0.set_xlabel('{:s} level'.format(label))
        ax1.set_xlabel('Resampled {:s} level'.format(label))

        return fig

    def show_resampling(self, xbins=None, ybins=None, **kwargs):
        """
        Visualize resampling procedure.

        Returns:

            figures (tuple)

        """

        # get raw background pixels
        if self.store_pixels:
            x0, y0 = self.xraw, self.yraw
        else:
            x0, y0 = self.extract_background()

        fig0 = self._pixel_distribution(x0, self.x, label='RFP', bins=xbins, **kwargs)
        fig1 = self._pixel_distribution(y0, self.y, label='GFP', bins=ybins, **kwargs)

        return fig0, fig1


class LayerCorrectionVisualization:
    """ Methods for visualizing layer correction procedure. """

    def show_background_extraction(self, **kwargs):
        """
        Visualize background extraction procedure.

        Keyword arguments:

            invert (bool) - if True, mask background rather than foreground

        Returns:

            figure

        """
        bg_extraction = BackgroundExtraction(self.layer, niters=self.niters)
        return bg_extraction.plot_foreground_mask(**kwargs)



from unittest.mock import patch
from os.path import join, exists

from flyqma.data import Experiment, Stack
from .test_io import TestPaths


class Test04_Experiment(TestPaths):
    """
    Tests for Experiment instance.
    """

    @classmethod
    def setUpClass(cls):
        """ Initialize test instance of Experiment. """
        with patch('builtins.input', side_effect=['yes', 12]):
            experiment = Experiment(cls.experiment_path)
        cls.experiment = experiment

    @classmethod
    def tearDownClass(cls):
        """ Restore stack directories to original state. """
        for stack in cls.experiment:
            stack.restore_directory()

    def test01_load_stack(self):
        """ Load Stack instance from Experiment. """
        stack = self.experiment.load_stack(self.experiment.stack_ids[0])
        self.assertTrue(isinstance(stack, Stack))



from os.path import join, exists
from os import mkdir
import shutil
import json
import numpy as np
import dill as pickle


def safeload(loader):
    """ Decorator for checking files exist before attempting to load them. """
    def wrapper(self, path):
        out = None
        if exists(path):
            out = loader(path)
        return out
    return wrapper


class IO:
    """
    Class provides assorted methods for reading/writing different filetypes.
    """

    @staticmethod
    def make_dir(dir_name, path=None, force=False):
        """ Create Silhouette3D container. """

        # compile directory name
        if path is not None:
            dir_name = join(path, dir_name)

        # check if directory exists
        if exists(dir_name):
            if force == True:
                shutil.rmtree(dir_name, ignore_errors=True)
            else:
                print('Will not overwrite existing directory.')
                return dir_name

        # make dir
        mkdir(dir_name)

        return dir_name

    @safeload
    def read_json(path):
        """ Read data from json. """
        with open(path, 'r') as f:
             data = json.load(f)
        return data

    @staticmethod
    def write_json(path, data):
        """ Write data to json. """
        with open(path, 'w') as f:
             json.dump(data, f, sort_keys=True, indent='\t')

    @safeload
    def read_tiff(path):
        """ Read stack from tiff """
        import tifffile as tf
        im = tf.imread(path)
        return im

    @staticmethod
    def render_tiff(path, stack):
        """ Render stack as tiff. """
        if stack is not None:
            tf.imsave(path, stack)

    @safeload
    def read_csv(path):
        import pandas as pd
        return pd.read_csv(path, index_col=0)

    @staticmethod
    def write_csv(path, data):
        if data is not None:
            data.to_csv(path)

    @safeload
    def read_npy(path):
        return np.load(path)

    @staticmethod
    def write_npy(path, arr):
        if arr is not None:
            np.save(path, arr)


class Pickler:
    """ Methods for pickling an object instance. """

    def save(self, filepath):
        """ Save serialized instance. """
        with open(filepath, 'wb') as file:
            pickle.dump(self, file, protocol=0)

    @staticmethod
    def load(filepath):
        """ Save serialized instance. """
        with open(filepath, 'rb') as file:
            batch = pickle.load(file)
        return batch



from os.path import join
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Polygon
from matplotlib.gridspec import GridSpec

from ..utilities import IO


class LayerVisualization:
    """
    Object for visualizing a single layer.

    Attributes:

        path (str) - layer path

        layer (flyqma.Layer) - layer instance

        axes (array like) - axes for blue/red/green color channels

    """

    def __init__(self, layer, axes):
        """
        Instantiate layer visualization.

        Args:

            layer (Layer) - image layer

            axes (array like) - axes for blue/red/green color channels

        """

        self.layer = layer

        # set selection path
        layer.make_subdir('selection')
        self.path = layer.subdirs['selection']

        # set axes
        self.axes = axes

        # render images
        self.render_images(layer)

    def render_images(self, layer, cmap=None):
        """
        Add blue, green, and red channels of layer to axes.

        Args:

            layer (Layer) - RGB image layer

            cmap (matplotlib.colors.ColorMap)

        """

        # visualize layers
        for ch, ax in enumerate(self.axes):
            _ = layer.get_channel(ch).show(segments=False, ax=ax, cmap=cmap)
            ax.set_aspect(1)

        # add layer number
        text = ' {:d}'.format(layer._id)
        self.axes[0].text(0, 0, text, fontsize=14, color='y', va='top')

    def add_marker(self, x, y, color='k', markersize=10):
        """ Add marker to layer images. """
        for ax in self.axes:
            ax.plot(x, y, '.', color=color, markersize=markersize, zorder=2)

    def remove_marker(self):
        """ Remove marker from layer images. """
        for ax in self.axes:
            ax.lines[-1].remove()

    def update_marker(self, color, markersize, ind=-1):
        """ Update size and color of last added marker. """
        for ax in self.axes:
            if len(ax.lines) > 0:
                ax.lines[ind].set_color(color)
                ax.lines[ind].set_markersize(markersize)

    def clear_markers(self):
        """" Remove all markers. """
        while len(self.axes[0].lines) > 0:
            self.remove_marker()

    def add_polygon(self):
        """ Add polygon to each image. """
        for ax in self.axes:
            poly = Polygon(self.pts,
                       ec=(1,1,0,1), lw=1,
                       fc=(1,1,1,0.2), fill=False,
                       zorder=1, closed=True)
            ax.add_patch(poly)

    def remove_polygon(self):
        """ Remove polygon from each image. """
        for ax in self.axes:
            ax.patches[0].remove()

    def overlay(self, msg, s=18):
        """ Overlay string centered on image. """
        for ax in self.axes:
            ax.images[0].set_alpha(0.5)
            x, y = np.mean(ax.get_xlim()), np.mean(ax.get_ylim())
            ax.text(x, y, msg, color='k', fontsize=s, ha='center', va='center')


class LayerInterface(LayerVisualization):
    """
    Event handler for an individual layer.


    Attributes:

        include (bool) - flag for layer inclusion

        active_polyhon (bool) - if True, polygon is currently active

        pts (list) - selection boundary points

        traceback (list) - exception traceback


    Inherited attributes:

        path (str) - layer path

        layer (flyqma.Layer) - layer instance

        axes (array like) - axes for blue/red/green color channels

    """

    def __init__(self, layer, axes):
        """
        Instantiate layer interface.

        Args:

            layer (Layer) - image layer

            axes (array like) - axes for blue/red/green color channels

        """

        # call visualization instantiation method
        super().__init__(layer, axes)

        # set layer attributes
        self.include = True

        # no initial polygon
        self.active_polygon = False

        # initialize points list
        self.pts = []
        self.traceback = []

    def load(self):
        """ Load layer selection. """
        io = IO()

        # load selected points
        pts = io.read_npy(join(self.path, 'selection.npy'))
        self.pts = pts.tolist()

        # load selection metadata
        md = io.read_json(join(self.path, 'md.json'))
        self.include = md['include']

        # add markers
        for pt in self.pts:
             self.add_marker(*pt, color='y', markersize=5)
        self.update_marker('r', markersize=10)

        # add polygon
        if len(self.pts) >= 3:
            self.add_polygon()
            self.active_polygon = True

        # mark excluded layers
        if self.include==False:
            self.overlay('EXCLUDED')

    def save(self):
        """ Save selected points and selection metadata to file. """

        # if no region was specified, use corners (e.g. include everything)
        if len(self.pts) <= 2:
            w,h = self.layer.shape
            self.pts = [ [0,0], [w, 0], [w, h], [0, h] ]

        io = IO()
        pts = np.array(self.pts)
        io.write_npy(join(self.path, 'selection.npy'), pts)
        md = dict(include=self.include)
        io.write_json(join(self.path, 'md.json'), md)

        # update measurements
        self.layer.load_inclusion()
        self.layer.define_roi(self.layer.data)
        self.layer.save_processed_data()

    def clear(self):
        """ Clear all points from layer selection bounds. """
        self.pts = []
        self.clear_markers()
        if len(self.axes[0].patches) > 0:
            self.remove_polygon()
            self.active_polygon = False

    def add_point(self, pt):
        """ Add point to layer selection bounds. """

        # store point
        self.pts.append(pt)

        # update previous marker and add new marker
        self.update_marker(color='y', markersize=5)
        self.add_marker(*pt, color='r', markersize=10)

        # update polygon
        if self.active_polygon:
            self.update_polygon()
        elif len(self.pts) == 3:
            self.add_polygon()
            self.active_polygon = True

    def remove_point(self):
        """ Remove last point added to layer selection bounds. """
        _ = self.pts.pop()
        self.remove_marker()
        self.update_marker(color='r', markersize=10)
        if len(self.pts) < 3:
            self.active_polygon = False

    def update_polygon(self):
        """ Update polygon for each image. """
        self.remove_polygon()
        if self.active_polygon:
            self.add_polygon()

    def undo(self):
        """ Remove last point added and update polygon. """
        self.remove_point()
        self.update_polygon()


class StackInterface:
    """
    Object for visualizing multiple layers in an image stack.

    Attributes:

        path (str) - layer path

        axes (array like) - axes for blue/red/green color channels

    """

    def __init__(self, stack):
        self.path = stack.path
        self.build_interface(stack)

    def build_interface(self, stack):
        """
        Build interface by adding interface for each layer.

        Args:

            stack (Stack) - image stack

        """

        # create figure
        nrows, ncols = stack.stack_depth, stack.color_depth
        figsize = (2.25 * ncols, 2.25 * nrows)
        self.fig = plt.figure(figsize=figsize)
        gs = GridSpec(nrows=nrows, ncols=ncols, wspace=.01, hspace=.01)

        # instantiate maps
        self.layer_to_interface = {}
        self.ax_to_layer = {}

        # build interface for each layer
        for i in range(stack.stack_depth):

            # iterator excludes discs marked for excludion
            layer = stack[i]

            # create all axes for current layer
            axes = [self.fig.add_subplot(gs[i*ncols+j]) for j in range(ncols)]

            # add layer gui to layer --> interface map
            self.layer_to_interface[i] = LayerInterface(layer, axes)

            # update axis --> layer map
            for ax in axes:
                self.ax_to_layer[ax] = i

            # label top row
            if i == 0:
                for j, ax in enumerate(axes):
                    ch_label = 'Channel {:d}'.format(j)
                    ax.set_title(ch_label, fontsize=14)



from os.path import join, exists
from os import mkdir
import gc
import numpy as np
from sklearn.cluster import k_means
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize

from ...utilities import IO


class ClassifierIO:
    """
    Methods for saving and loading classifier objects.
    """

    def save(self, dirpath, data=False, image=True, extension=None, **kwargs):
        """
        Save classifier to specified path.

        Args:

            dirpath (str) - directory in which classifier is to be saved

            data (bool) - if True, save training data

            image (bool) - if True, save labeled histogram image

            extension (str) - directory name extension

            kwargs: keyword arguments for image rendering

        """

        # create directory for classifier
        dirname = 'classifier'
        if extension is not None:
            dirname += '_{:s}'.format(str(extension))
        path = join(dirpath, dirname)
        if not exists(path):
            mkdir(path)

        # save values
        if data:
            np.save(join(path, 'values.npy'), self._values)

        # save parameters
        io = IO()
        io.write_json(join(path, 'parameters.json'), self.parameters)

        # save image
        if image:
            pass

            # # visualize classification
            # self.show()

            # # save image
            # self.fig.savefig(join(path, 'classifier.pdf'), **kwargs)
            # self.fig.clf()
            # plt.close(self.fig)
            # gc.collect()

        return path

    @classmethod
    def load(cls, path):
        """
        Load classifier from file.

        Args:

            path (str) - path to classifier directory

        Returns:

            classifier (Classifier derivative)

        """
        io = IO()

        values_path = join(path, 'values.npy')
        if exists(values_path):
            values = io.read_npy(values_path)
        else:
            values = None

        parameters = io.read_json(join(path, 'parameters.json'))
        return cls(values, **parameters)


class ClassifierProperties:
    """
    Properties for classifier objects.
    """

    @property
    def num_samples(self):
        """ Number of samples. """
        return len(self._values)

    @property
    def values(self):
        """ Values for classifier. """
        if self.log:
            return np.log(self._values)
        else:
            return self._values

    @property
    def order(self):
        """ Ordered component indices (low to high). """
        x = self.component_to_label
        return sorted(x, key=x.__getitem__)

    @property
    def component_groups(self):
        """ List of lists of components for each label. """
        x = self.component_to_label
        labels = np.unique(list(x.values()))
        return [[k for k, v in x.items() if v == l] for l in labels]

    @property
    def centroids(self):
        """ Means of each component (not log transformed). """
        centroids = self.model.means_
        if self.log:
            centroids = np.exp(centroids)
        return centroids

    @property
    def component_to_label(self):
        """
        Returns dictionary mapping components to labels.  Mapping is achieved by k-means clustering the model centroids (linear scale).
        """
        n = self.num_labels
        cluster_means, cluster_labels, _ = k_means(self.centroids, n, n_init=100)
        component_to_label = {}
        for label, c in enumerate(np.argsort(cluster_means.mean(axis=1))):
            for d in (cluster_labels==c).nonzero()[0]:
                component_to_label[d] = label
        return component_to_label


class Classifier(ClassifierProperties, ClassifierIO):
    """
    Classifier base class. Children of this class must possess a means attribute, as well as a predict method.


    Attributes:

        values (array like) - basis for clustering

        attribute (str or list) - attribute(s) used to determine labels

        log (bool) - indicates whether clustering performed on log values

        num_labels (int) - number of output labels

        classifier (vectorized func) - maps value to label_id

        labels (np.ndarray[int]) - predicted labels

        cmap (matplotlib.colors.ColorMap) - colormap for label_id

        parameters (dict) - {param name: param value} pairs

        fig (matplotlib.figures.Figure) - histogram figure

    """

    def __init__(self, values,
                 attribute=None,
                 num_labels=3,
                 log=True,
                 cmap=None):
        """
        Instantiate classifier mapping <n> clusters to <num_labels>.

        Args:

            values (np.ndarray[float]) - basis for clustering

            attribute (str or list) - attribute(s) on which to cluster

            num_labels (int) - number of class labels

            log (bool) - if True, cluster log-transformed values

            cmap (matplotlib.colors.ColorMap) - colormap for cell labels

        """

        # set values, whether to log transform them, and number of clusters
        self._values = values
        self.log = log

        self.num_labels = num_labels

        # set colormap
        self.set_cmap(cmap=cmap)

        # store parameters
        if type(attribute) == str:
            attribute = [attribute]
        self.attribute = attribute
        self.parameters = dict(num_labels=num_labels,
                               log=log,
                               attribute=attribute)
        self.fig = None

    def __call__(self, data):
        """
        Assign class labels to <data>.

        Args:

            data (pd.DataFrame) - must contain necessary attributes

        Returns:

            labels (np.ndarray[int])

        """
        return self.evaluate_classifier(data)

    def evaluate_classifier(self, data):
        """
        Assign class labels to <data>.

        Args:

            data (pd.DataFrame) - must contain necessary attributes

        Returns:

            labels (np.ndarray[int])

        """
        x =  data[self.attribute].values
        if self.log:
            x = np.log(x)
        return self.classifier(x)

    @classmethod
    def from_measurements(cls, data, attribute, **kwargs):
        """
        Fit classifier to data.

        Args:

            data (pd.DataFrame) - measurement data

            attribute (str or list) - attribute(s) on which to cluster

            kwargs: keyword arguments for classifier

        Returns:

            classifier (Classifier derivative)

        """
        return cls(data[attribute].values, attribute, **kwargs)

    @classmethod
    def from_grouped_measurements(cls,
                            data,
                            attribute,
                            groupby=None,
                            **kwargs):
        """
        Fit classifier to data grouped by a specified feature.

        Args:

            data (pd.DataFrame) - measurement data

            groupby (str) - attribute used to group measurement data

            attribute (str or list) - attribute(s) on which to cluster

            kwargs: keyword arguments for classifier

        Returns:

            classifier (Classifier derivative)

        """

        if groupby is None:
            groupby = ('disc_genotype', 'disc_id', 'layer', 'im_label')
        values = data.groupby(by=groupby)[attribute].mean().values
        return cls(values, attribute, **kwargs)

    def show(self):
        """ Visualize classification. """
        pass

    def set_cmap(self, cmap=None, vmin=0, vmax=None):
        """
        Set colormap for class labels.

        Args:

            cmap (matplotlib.colormap)

            vmin (int) - lower bound for color scale

            vmax (int) - upper bound for color scale

        """

        # select colormap
        if cmap is None:
            cmap = plt.cm.viridis

        if vmax is None:
            vmax = self.num_labels - 1

        # normalize
        norm = Normalize(vmin=vmin, vmax=vmax)
        self.cmap = lambda x: cmap(norm(x))

    def build_colormap(self, cmap, vmin=-1):
        """
        Build normalized colormap for class labels.

        Args:

            cmap (matplotlib.colormap)

            vmin (float) - lower bound for colorscale

        Returns:

            colormap (func) - function mapping class labels to colors

        """
        norm = Normalize(vmin=vmin, vmax=self.num_labels-1)
        return lambda x: cmap(norm(x))

    def build_classifier(self):
        """
        Build function that returns the most probable label for each of a series of values.
        """

        # build classifier that maps model components to labels.
        component_to_label = np.vectorize(self.component_to_label.get)

        def classifier(values):
            """ Returns <label> for <values>.  """
            return component_to_label(self.predict(values))

        return classifier



from os.path import join, abspath, exists
import json
import pandas as pd


class ReadSilhouette:
    """

    Read-only interface to a FlyEye Silhouette file.

    Attributes:

        path (str) - path to Silhouette file

        feed (dict) - feed file containing layer IDs

        feud (dict) - feud file containing cell type labels

    Properties:

        is_flipped_about_yz (bool) - if True, invert about YZ plane

        is_flipped_about_xy (bool) - if True, invert about XY plane

    """
    def __init__(self, path):
        """
        Instantiate interface to silhouette file.

        Args:

            path (str) - path to silhouette file

        """
        self.path = abspath(path)

        # load feed and feud files
        self.feed = self.read_json('feed.json')
        self.feud = self.read_json('feud.json')

    def read_json(self, filename):
        """
        Read contents of specified JSON file.

        Args:

            filename (str) - filename

        Returns:

            out (dict) - file contents

        """
        filepath = join(self.path, filename)
        with open(filepath, 'r') as f:
            out = json.load(f)
        return out


class ReadSilhouetteData(ReadSilhouette):
    """

    Read-only interface to data within a FlyEye Silhouette file.

    Upon instantiation, individual cell measurements are aggregated into a data.cells.Cells compatible DataFrame.

    Measurement data must be read on a layer-by-layer basis the first time a Silhouette object is instantiated. Following this initial reading, the aggregated measurement data are serialized and stored within the silhouette file. These serialized measurements may then be accessed directly during future use. The recompile flag indicates whether the serialized measurements should be ignored upon instantiation.

    Attributes:

        df (pd.DataFrame) - cell measurement data

    Inherited attributes:

        path (str) - path to Silhouette file

        feed (dict) - feed file containing layer IDs

        feud (dict) - feud file containing cell type labels

    """

    def __init__(self, path, recompile=False):
        """
        Instantiate interface to silhouette file data.

        Args:

            path (str) - path to silhouette file

            recompile (bool) - if True, recompile measurements from all layers

        """
        super().__init__(path)
        self.load(recompile=recompile)

    @property
    def labels(self):
        """ pd.Series of labels keyed by (layer_id, segment_id). """
        return self.df.set_index(['layer', 'segment_id'])['label']

    def compile_measurements(self):
        """ Compile measurements from all layers (slow access). """
        labels = self.read_labels()
        self.df = self.read_contours(labels)

    def save_measurements(self):
        """ Save serialized measurements for fast access. """
        self.df.to_json(join(self.path, 'measurements.json'))

    def load_measurements(self):
        """ Load serialized measurements (fast access). """
        self.df = pd.read_json(join(self.path, 'measurements.json'))

    def load(self, recompile=False):
        """
        Read all contour and orientation data from silhouette file.

        Args:

            recompile (bool) - if True, recompile measurements from all layers

        """

        # check whether measurements are available
        measurements_available = exists(join(self.path, 'measurements.json'))

        # load available measurements if recompile flag is false
        if measurements_available and not recompile:
            self.load_measurements()

        # otherwise, recompile and save measurements
        else:
            self.compile_measurements()
            self.save_measurements()

    def read_labels(self):
        """
        Load segment labels from silhouette file.

        Returns:

            labels (dict) - {layer_id: {contour_id: label}} entries for each layer

        """

        # compile labels for all layers
        labels = {}
        for layer in self.feud['layers']:

            # compile {contour_id: contour_label} dictionary for current layer
            annotations = {}
            for contour in layer['contours']:
                label = contour.get('label', None)
                if label is None or label.strip() == '':
                    continue
                annotations[contour['id']] = contour['label']

            # store labels for current layer
            labels[layer['id']] = annotations

        return labels

    @staticmethod
    def parse_contour(contour):
        """
        Convert contour to list format.

        Args:

            contour (dict) - contour from silhouette file

        Returns:

            ctr_list (list) - values in data.cells.Cells compatible list format

        """

        # extract values
        centroid = contour['centroid']
        pixel_count = contour['pixel_count']
        segment = contour['points']
        color_avg = contour['color_avg']
        color_std = contour['color_std']

        # reorganize values
        ctr_list = [contour['id'],
                    centroid[0], centroid[1],
                    segment,
                    pixel_count]

        # parse measurements
        keys1, avgs = list(zip(*color_avg.items()))
        keys2, stds = list(zip(*color_std.items()))
        ctr_list.extend(avgs)
        ctr_list.extend(stds)
        assert sum([k1!=k2 for k1,k2 in zip(keys1,keys2)])==0, 'Contour keys do not match.'

        return keys1, ctr_list

    def read_contours(self, all_labels={}, include_unlabeled=False):
        """
        Read contours from silhouette file.

        Args:

            all_labels (dict) - {layer_id: {contour_id: label}} for each layer

            include_unlabeled (bool) - if True, include unlabeled segments

        Returns:

            df (pd.DataFrame) - data.cells.Cells compatible dataframe of contours

        """

        # read contours from all layers
        contours = []
        for layer_id in self.feed['layer_ids']:

            # load labels for current layer
            labels = all_labels.get(layer_id, None)

            # skip layers without any labels
            if labels is None and not include_unlabeled:
                continue

            # read layer from silhouette file
            layer = self.read_json('{:d}.json'.format(layer_id))

            # read all contours within layer
            for contour in layer['contours']:

                # get label for current contour
                if labels is None:
                    label = None
                else:
                    label = labels.get(contour['id'], None)

                if label is None and not include_unlabeled:
                    continue

                # convert to list format
                keys, ctr_list = self.parse_contour(contour)
                ctr_list.extend([layer_id, label])

                # store contours from current layer
                contours.append(ctr_list)

        # compile dataframe
        columns = ['segment_id',
                   'centroid_x',
                   'centroid_y',
                   'segment',
                   'pixel_count']
        columns += keys
        columns += ['{:s}_std'.format(k) for k in keys]
        columns += ['layer', 'label']
        df = pd.DataFrame(contours, columns=columns)

        # delete duplicate RGB channel labels
        for i, c in enumerate('rgb'):
            if 'ch{:d}'.format(i) in df.columns:
                df.drop(c, axis=1, inplace=True)
                df.drop(c+'_std', axis=1, inplace=True)

        return df



import numpy as np
from sklearn.cluster import KMeans

from .classifiers import Classifier


class KMeansClassifier(Classifier):
    """
    K-means classifier.

    Attributes:

        groups (dict) - {cluster_id: label_id} pairs for merging clusters

        component_to_label (vectorized func) - maps cluster_id to label_id

        km (sklearn.cluster.KMeans) - kmeans object

        classifier (vectorized func) - maps value to label_id

        labels (np.ndarray[int]) - predicted labels

    Inherited attributes:

        values (array like) - basis for clustering

        attribute (str or list) - attribute(s) on which to cluster

        log (bool) - indicates whether clustering performed on log values

        cmap (matplotlib.colors.ColorMap) - colormap for label_id

        parameters (dict) - {param name: param value} pairs

        fig (matplotlib.figures.Figure) - histogram figure

    """

    def __init__(self, values,
                 num_components=3,
                 groups=None,
                 log=True,
                 **kwargs):
        """
        Instantiate k-means classifier.

        Args:

            values (array like) - basis for clustering

            num_components (int) - number of clusters

            groups (dict) - {cluster_id: label_id} pairs for merging clusters

            log (bool) - indicates whether clustering performed on log values

            kwargs: keyword arguments for Classifier parent class

        """

        # set groups and number of clusters
        if groups is None:
            groups = {k: k for k in range(num_components)}
        else:
            groups = {int(k): v for k, v in groups.items()}
        num_labels = len(groups)

        # instantiate classifier
        super().__init__(values, num_labels=num_labels, log=log, **kwargs)
        self.num_components = num_components
        self.component_to_label = np.vectorize(groups.get)
        self.groups = groups

        # build classifiers
        self.model = self.fit(self.values, self.num_components)
        self.classifier = self._build_value_to_groups_classifier()

        # assign group labels
        self.labels = self.classifier(self.values.reshape(-1, 1))

        # store parameters
        self.parameters.update(dict(groups=self.groups))

    @property
    def means(self):
        """ Mean of each cluster. """
        return self.model.cluster_centers_.ravel()

    def predict(self, values):
        """ Predict which component each of <values> belongs to. """
        return self.model.predict(values)

    @staticmethod
    def fit(values, n):
        """ Fit n clusters to x """
        return KMeans(n).fit(values.reshape(-1, 1))

    @staticmethod
    def _build_value_to_cluster_classifier(km):
        """ Build classifier mapping values to sequential clusters. """
        centroids = km.cluster_centers_.ravel()
        flip = lambda f: f.__class__(map(reversed, f.items()))
        km_to_ordered_dict = flip(dict(enumerate(np.argsort(centroids))))
        km_to_ordered = np.vectorize(km_to_ordered_dict.get)
        classifier = lambda x: km_to_ordered(km.predict(x))
        return classifier

    def _build_value_to_groups_classifier(self):
        """ Build classifier mapping values to groups. """
        value_to_cluster = self._build_value_to_cluster_classifier(self.model)
        classifier = lambda x: self.component_to_label(value_to_cluster(x))
        return classifier
